# -*- coding: utf-8 -*-
"""
Test the Recorder object API

"""
import pywr.core
from pywr.core import Model, Input, Output, Scenario, AggregatedNode
import numpy as np
import pandas
import pytest
import tables
import json
from numpy.testing import assert_allclose, assert_equal
from fixtures import simple_linear_model, simple_storage_model
from pywr.recorders import (Recorder, NumpyArrayNodeRecorder, NumpyArrayStorageRecorder, NumpyArrayAreaRecorder, NumpyArrayLevelRecorder,
                            AggregatedRecorder, CSVRecorder, TablesRecorder, TotalDeficitNodeRecorder,
                            TotalFlowNodeRecorder, RollingMeanFlowNodeRecorder, MeanFlowNodeRecorder, NumpyArrayParameterRecorder,
                            NumpyArrayIndexParameterRecorder, RollingWindowParameterRecorder, AnnualCountIndexParameterRecorder,
                            RootMeanSquaredErrorNodeRecorder, MeanAbsoluteErrorNodeRecorder, MeanSquareErrorNodeRecorder,
                            PercentBiasNodeRecorder, RMSEStandardDeviationRatioNodeRecorder, NashSutcliffeEfficiencyNodeRecorder,
                            EventRecorder, Event, StorageThresholdRecorder, NodeThresholdRecorder, EventDurationRecorder, EventStatisticRecorder,
                            FlowDurationCurveRecorder, FlowDurationCurveDeviationRecorder, StorageDurationCurveRecorder,
                            HydropowerRecorder, TotalHydroEnergyRecorder,
                            TotalParameterRecorder, MeanParameterRecorder, NumpyArrayNodeCostRecorder,
                            NumpyArrayNodeDeficitRecorder, NumpyArrayNodeSuppliedRatioRecorder, NumpyArrayNodeCurtailmentRatioRecorder,
                            SeasonalFlowDurationCurveRecorder, load_recorder, ParameterNameWarning, NumpyArrayDailyProfileParameterRecorder,
                            AnnualTotalFlowRecorder, AnnualCountIndexThresholdRecorder, TimestepCountIndexParameterRecorder,
                            GaussianKDEStorageRecorder, NormalisedGaussianKDEStorageRecorder, NumpyArrayNormalisedStorageRecorder
                            )

from pywr.recorders.progress import ProgressRecorder

from pywr.parameters import (DailyProfileParameter, FunctionParameter, ArrayIndexedParameter, ConstantParameter,
                             InterpolatedVolumeParameter, ConstantScenarioParameter, StorageThresholdParameter)
from helpers import load_model
import os
import sys


class TestRecorder:
    """Tests for Recorder base class."""

    def test_default_no_constraint(self, simple_linear_model):
        """Test the constraint properties in the default instance (i.e. not a constraint)."""
        r = Recorder(simple_linear_model)
        assert r.constraint_lower_bounds is None
        assert r.constraint_upper_bounds is None
        assert not r.is_constraint
        assert not r.is_lower_bounded_constraint
        assert not r.is_double_bounded_constraint
        assert not r.is_upper_bounded_constraint
        assert not r.is_equality_constraint

    def test_equality_constraint(self, simple_linear_model):
        """Test equality constraint identification. """
        r = Recorder(simple_linear_model, constraint_lower_bounds=10.0, constraint_upper_bounds=10.0)
        assert r.is_constraint
        assert not r.is_lower_bounded_constraint
        assert not r.is_double_bounded_constraint
        assert not r.is_upper_bounded_constraint
        assert r.is_equality_constraint

    def test_lower_bounded_constraint(self, simple_linear_model):
        """Test lower bounded constraint identification. """
        r = Recorder(simple_linear_model, constraint_lower_bounds=10.0, constraint_upper_bounds=None)
        assert r.is_constraint
        assert r.is_lower_bounded_constraint
        assert not r.is_double_bounded_constraint
        assert not r.is_upper_bounded_constraint
        assert not r.is_equality_constraint

    def test_upper_bounded_constraint(self, simple_linear_model):
        """Test upper bounded constraint identification. """
        r = Recorder(simple_linear_model, constraint_lower_bounds=None, constraint_upper_bounds=10.0)
        assert r.is_constraint
        assert not r.is_lower_bounded_constraint
        assert not r.is_double_bounded_constraint
        assert r.is_upper_bounded_constraint
        assert not r.is_equality_constraint

    def test_double_bounded_constraint(self, simple_linear_model):
        """Test upper bounds constraint identification. """
        r = Recorder(simple_linear_model, constraint_lower_bounds=2.0, constraint_upper_bounds=10.0)
        assert r.is_constraint
        assert not r.is_lower_bounded_constraint
        assert r.is_double_bounded_constraint
        assert not r.is_upper_bounded_constraint
        assert not r.is_equality_constraint

    def test_invalid_bounds_constraint(self, simple_linear_model):
        """Test lower bounds greater than upper bounds."""
        with pytest.raises(ValueError):
            r = Recorder(simple_linear_model, constraint_lower_bounds=10.0, constraint_upper_bounds=2.0)

    @pytest.mark.parametrize('lb, ub', (
            (None, None),
            (5.0, 10.0),  # Feasible double bounds
            (15.0, 20.0),  # Infeasible double bounds
            (0.0, 2.0),  # Infeasible double bounds
            (0.0, None),  # Feasible lower bounds
            (15.0, None),  # Infeasible lower bounds
            (None, 15.0),  # Feasible upper bounds
            (None, 5.0),  # Infeasible upper bounds
    ))
    def test_is_constraint_violated(self, simple_linear_model, lb, ub):
        """Test the calculation of a violated constraint and model feasibility."""
        m = simple_linear_model

        class TestRecorder(Recorder):
            def aggregated_value(self):
                return 10.0

        r = TestRecorder(m, constraint_lower_bounds=lb, constraint_upper_bounds=ub)

        if lb is None and ub is None:
            assert not r.is_constraint
            with pytest.raises(ValueError):
                r.is_constraint_violated()
        elif lb is None and ub is not None:
            # Upper bounded only
            if ub >= 10.0:
                assert not r.is_constraint_violated()
                assert m.is_feasible()
            else:
                assert r.is_constraint_violated()
                assert not m.is_feasible()
        elif lb is not None and ub is None:
            # Lower bounded only
            if lb <= 10.0:
                assert not r.is_constraint_violated()
                assert m.is_feasible()
            else:
                assert r.is_constraint_violated()
                assert not m.is_feasible()
        else:
            # Double bounds
            if lb <= 10.0 <= ub:
                assert not r.is_constraint_violated()
                assert m.is_feasible()
            else:
                assert r.is_constraint_violated()
                assert not m.is_feasible()


def test_numpy_recorder(simple_linear_model):
    """
    Test the NumpyArrayNodeRecorder
    """
    model = simple_linear_model
    otpt = model.nodes['Output']

    model.nodes['Input'].max_flow = 10.0
    otpt.cost = -2.0
    rec = NumpyArrayNodeRecorder(model, otpt)

    # test retrieval of recorder
    assert model.recorders['numpyarraynoderecorder.Output'] == rec
    # test changing name of recorder
    rec.name = 'timeseries.Output'
    assert model.recorders['timeseries.Output'] == rec
    with pytest.raises(KeyError):
        model.recorders['numpyarraynoderecorder.Output']

    model.run()

    assert rec.data.shape == (365, 1)
    assert np.all((rec.data - 10.0) < 1e-12)

    df = rec.to_dataframe()
    assert df.shape == (365, 1)
    assert np.all((df.values - 10.0) < 1e-12)


def test_numpy_recorder_from_json(simple_linear_model):
    """ Test loading NumpyArrayNodeRecorder from JSON style data """

    model = simple_linear_model

    data = {
        "type": "numpyarraynode",
        "node": "Output"
    }

    rec = load_recorder(model, data)
    assert isinstance(rec, NumpyArrayNodeRecorder)

def test_numpy_recorder_factored(simple_linear_model):
    """Test the optional factor applies correctly """

    model = simple_linear_model
    otpt = model.nodes['Output']
    otpt.max_flow = 30.0
    model.nodes['Input'].max_flow = 10.0
    otpt.cost = -2

    factor = 2.0
    rec_fact = NumpyArrayNodeRecorder(model, otpt, factor=factor)

    model.run()

    assert rec_fact.data.shape == (365, 1)
    assert_allclose(20, rec_fact.data, atol=1e-7)


class TestNumpyArrayCostRecorder:
    def test_simple(self, simple_linear_model):
        """Test NumpyArrayCostRecorder with fixed cost"""
        model = simple_linear_model
        model.nodes['Input'].max_flow = 10.0
        otpt = model.nodes['Output']
        otpt.cost = -2.0

        cost_rec = NumpyArrayNodeCostRecorder(model, otpt)
        # test retrieval of recorder
        assert model.recorders['numpyarraynodecostrecorder.Output'] == cost_rec
        model.run()

        assert cost_rec.data.shape == (365, 1)
        np.testing.assert_allclose(cost_rec.data, np.ones_like(cost_rec.data) * -2.0)

        df = cost_rec.to_dataframe()
        assert df.shape == (365, 1)
        np.testing.assert_allclose(df.values, np.ones_like(cost_rec.data) * -2.0)

    def test_scenarios(self, simple_linear_model):
        """Test NumpyArrayNodeCostRecorder with scenario varying costs."""
        model = simple_linear_model
        model.nodes['Input'].max_flow = 10.0
        otpt = model.nodes['Output']
        scenario = Scenario(model, name='A', size=2)
        otpt.cost = ConstantScenarioParameter(model, values=[-2, -10], scenario=scenario)

        cost_rec = NumpyArrayNodeCostRecorder(model, otpt)
        # test retrieval of recorder
        assert model.recorders['numpyarraynodecostrecorder.Output'] == cost_rec
        model.run()

        assert cost_rec.data.shape == (365, 2)
        expected = np.empty_like(cost_rec.data)
        expected[:, 0] = -2.0
        expected[:, 1] = -10.0
        np.testing.assert_allclose(cost_rec.data, expected)

        df = cost_rec.to_dataframe()
        assert df.shape == (365, 2)
        np.testing.assert_allclose(df.values, expected)

    def test_time_varying(self, simple_linear_model):
        """Test NumpyArrayNodeCostRecorder with time varying costs."""
        model = simple_linear_model
        model.nodes['Input'].max_flow = 10.0
        otpt = model.nodes['Output']

        costs = -2.0 - np.arange(365) * 0.1
        otpt.cost = ArrayIndexedParameter(model, costs)

        cost_rec = NumpyArrayNodeCostRecorder(model, otpt)
        # test retrieval of recorder
        assert model.recorders['numpyarraynodecostrecorder.Output'] == cost_rec
        model.run()

        assert cost_rec.data.shape == (365, 1)
        np.testing.assert_allclose(cost_rec.data, costs[:, np.newaxis])

        df = cost_rec.to_dataframe()
        assert df.shape == (365, 1)
        np.testing.assert_allclose(df.values, costs[:, np.newaxis])


class TestFlowDurationCurveRecorders:
    funcs = {"min": np.min, "max": np.max, "mean": np.mean, "sum": np.sum}

    @pytest.mark.parametrize("agg_func", ["min", "max", "mean", "sum"])
    def test_fdc_recorder(self, agg_func):
        """
        Test the FlowDurationCurveRecorder
        """
        model = load_model("timeseries2.json")
        input = model.nodes['catchment1']

        percentiles = np.linspace(20., 100., 5)
        rec = FlowDurationCurveRecorder(model, input, percentiles, temporal_agg_func=agg_func, agg_func="min")

        # test retrieval of recorder
        assert model.recorders['flowdurationcurverecorder.catchment1'] == rec
        # test changing name of recorder
        rec.name = 'timeseries.Input'
        assert model.recorders['timeseries.Input'] == rec
        with pytest.raises(KeyError):
            model.recorders['flowdurationcurverecorder.catchment1']

        model.run()

        func = TestAggregatedRecorder.funcs[agg_func]

        assert_allclose(rec.fdc[:, 0], [20.42,  21.78,  23.22,  26.47,  29.31])
        assert_allclose(func(rec.fdc, axis=0), rec.values())
        assert_allclose(np.min(func(rec.fdc, axis=0)), rec.aggregated_value())

        assert rec.fdc.shape == (len(percentiles), len(model.scenarios.combinations))
        df = rec.to_dataframe()
        assert df.shape == (len(percentiles), len(model.scenarios.combinations))

    def test_seasonal_fdc_recorder(self):
        """
        Test the FlowDurationCurveRecorder
        """
        model = load_model("timeseries4.json")

        df = pandas.read_csv(os.path.join(os.path.dirname(__file__), 'models', 'timeseries3.csv'),
                             parse_dates=True, dayfirst=True, index_col=0)

        percentiles = np.linspace(20., 100., 5)

        summer_flows = df.loc[pandas.Timestamp("2014-06-01"):pandas.Timestamp("2014-08-31"), :]
        summer_fdc = np.percentile(summer_flows, percentiles, axis=0)

        model.run()

        rec = model.recorders["seasonal_fdc"]
        assert_allclose(rec.fdc, summer_fdc)

    @pytest.mark.parametrize("agg_func, aggregate", [
        ("min", False), ("max", False), ("mean", False), ("mean", True), ("sum", False)])
    def test_fdc_dev_recorder(self, agg_func, aggregate):
        """
        Test the FlowDurationCurveDeviationRecorder
        """
        model = load_model("timeseries2.json")
        input = model.nodes['catchment1']
        term = model.nodes['term1']
        scenarioA = model.scenarios['scenario A']

        natural_flow = pandas.read_csv(os.path.join(os.path.dirname(__file__), 'models', 'timeseries2.csv'),
                                       parse_dates=True, dayfirst=True, index_col=0)
        percentiles = np.linspace(20., 100., 5)

        natural_fdc = np.percentile(natural_flow, percentiles, axis=0)

        # Lower target is 20% below natural
        lower_input_fdc = natural_fdc * 0.8
        # Upper is 10% above
        upper_input_fdc = natural_fdc * 1.1

        if aggregate:
            # Setup only a single target for all scenarios.
            lower_input_fdc = lower_input_fdc.mean(axis=1)
            upper_input_fdc = upper_input_fdc.mean(axis=1)
            scenarioA = None

        rec = FlowDurationCurveDeviationRecorder.load(model, {
            "node": "term1",
            "percentiles": percentiles,
            "lower_target_fdc": lower_input_fdc,
            "upper_target_fdc": upper_input_fdc,
            "temporal_agg_func": agg_func,
            "agg_func": "mean",
            "scenario": "scenario A"
        })

        # test retrieval of recorder
        assert model.recorders['flowdurationcurvedeviationrecorder.term1'] == rec
        # test changing name of recorder
        rec.name = 'timeseries.Input'
        assert model.recorders['timeseries.Input'] == rec
        with pytest.raises(KeyError):
            model.recorders['flowdurationcurvedeviationrecorder.term1']

        model.run()

        actual_fdc = np.maximum(natural_fdc - 23, 0.0)

        if aggregate:
            lower_input_fdc = lower_input_fdc[:, np.newaxis]
            upper_input_fdc = upper_input_fdc[:, np.newaxis]

        # Compute deviation
        lower_deviation = (lower_input_fdc - actual_fdc) / lower_input_fdc
        upper_deviation = (actual_fdc - upper_input_fdc) / upper_input_fdc
        deviation = np.maximum(np.maximum(lower_deviation, upper_deviation), np.zeros_like(lower_deviation))

        func = TestAggregatedRecorder.funcs[agg_func]

        assert_allclose(rec.fdc_deviations[:, 0], deviation[:, 0])
        assert_allclose(func(rec.fdc_deviations, axis=0), rec.values())
        assert_allclose(np.mean(func(rec.fdc_deviations, axis=0)), rec.aggregated_value())

        assert rec.fdc_deviations.shape == (len(percentiles), len(model.scenarios.combinations))
        df = rec.to_dataframe()
        assert df.shape == (len(percentiles), len(model.scenarios.combinations))

    def test_deviation_single_target_lower(self):
        """Test deviation recorder with a lower target and no upper target"""

        model = load_model("timeseries2.json")
        input = model.nodes['catchment1']
        term = model.nodes['term1']
        scenarioA = model.scenarios['scenario A']

        natural_flow = pandas.read_csv(os.path.join(os.path.dirname(__file__), 'models', 'timeseries2.csv'),
                                       parse_dates=True, dayfirst=True, index_col=0)
        percentiles = np.linspace(20., 100., 5)

        natural_fdc = np.percentile(natural_flow, percentiles, axis=0)

        # Lower target is 20% below natural
        lower_input_fdc = natural_fdc * 0.8

        rec = FlowDurationCurveDeviationRecorder(model, term, percentiles, lower_target_fdc=lower_input_fdc,
                                                 scenario=scenarioA)

        model.run()

        actual_fdc = np.maximum(natural_fdc - 23, 0.0)

        lower_deviation = (lower_input_fdc - actual_fdc) / lower_input_fdc
        deviation = np.maximum(lower_deviation, np.zeros_like(lower_deviation))
        assert_allclose(rec.fdc_deviations[:, 0], deviation[:, 0])

    def test_deviation_single_target_upper(self):
        """Test deviation recorder with an upper target and no lower target"""

        model = load_model("timeseries2.json")
        input = model.nodes['catchment1']
        term = model.nodes['term1']
        scenarioA = model.scenarios['scenario A']

        natural_flow = pandas.read_csv(os.path.join(os.path.dirname(__file__), 'models', 'timeseries2.csv'),
                                       parse_dates=True, dayfirst=True, index_col=0)
        percentiles = np.linspace(20., 100., 5)

        natural_fdc = np.percentile(natural_flow, percentiles, axis=0)

        # Upper is 10% above
        upper_input_fdc = natural_fdc * 1.1

        rec = FlowDurationCurveDeviationRecorder(model, term, percentiles, upper_target_fdc=upper_input_fdc,
                                                 scenario=scenarioA)

        model.run()

        actual_fdc = np.maximum(natural_fdc - 23, 0.0)

        upper_deviation = (actual_fdc - upper_input_fdc) / upper_input_fdc
        deviation = np.maximum(upper_deviation, np.zeros_like(upper_deviation))
        assert_allclose(rec.fdc_deviations[:, 0], deviation[:, 0])

    def test_fdc_dev_from_json(self):
        """Test loading deviation recorder from json"""

        model = load_model("timeseries2_with_fdc.json")
        model.run()

        rec = model.recorders['fdc_dev1']
        df = rec.to_dataframe()
        assert df.shape == (5, len(model.scenarios.combinations))

        rec = model.recorders['fdc_dev2']
        df = rec.to_dataframe()
        assert df.shape == (5, len(model.scenarios.combinations))


def test_sdc_recorder():
    """
    Test the StorageDurationCurveRecorder
    """
    model = load_model("timeseries3.json")
    inpt = model.nodes['catchment1']
    strg = model.nodes['reservoir1']

    percentiles = np.linspace(20., 100., 5)
    flow_rec = NumpyArrayNodeRecorder(model, inpt)
    rec = StorageDurationCurveRecorder(model, strg, percentiles, temporal_agg_func="max", agg_func="min")

    # test retrieval of recorder
    assert model.recorders['storagedurationcurverecorder.reservoir1'] == rec

    model.run()

    # Manually calculate expected storage and percentiles
    strg_volume = strg.initial_volume + np.cumsum(flow_rec.data - 23.0, axis=0)
    strg_pciles = np.percentile(strg_volume, percentiles, axis=0)

    assert_allclose(rec.sdc, strg_pciles)
    assert_allclose(np.max(rec.sdc, axis=0), rec.values())
    assert_allclose(np.min(np.max(rec.sdc, axis=0)), rec.aggregated_value())

    assert rec.sdc.shape == (len(percentiles), len(model.scenarios.combinations))
    df = rec.to_dataframe()
    assert df.shape == (len(percentiles), len(model.scenarios.combinations))


@pytest.mark.parametrize('proportional', [True, False])
def test_numpy_storage_recorder(simple_storage_model, proportional):
    """
    Test the NumpyArrayStorageRecorder
    """
    model = simple_storage_model

    res = model.nodes['Storage']

    rec = NumpyArrayStorageRecorder(model, res, proportional=proportional)

    model.run()

    expected = np.array([[7, 4, 1, 0, 0]]).T
    if proportional:
        expected = expected / 20

    assert(rec.data.shape == (5, 1))
    assert_allclose(rec.data, expected, atol=1e-7)

    df = rec.to_dataframe()
    assert df.shape == (5, 1)
    assert_allclose(df.values, expected, atol=1e-7)


def test_numpy_norm_storage_recorder(simple_storage_model):
    """
    Test the NumpyArrayNormalisedStorageRecorder
    """
    model = simple_storage_model

    res = model.nodes['Storage']
    cc = ConstantParameter(model, value=0.2)

    rec = NumpyArrayNormalisedStorageRecorder(model, res, parameter=cc)

    model.run()

    expected = np.array([[
        3.0/16,  # 3 above control curve
        0,  # at control curve
        1.0/4 - 1.0,  # 75% toward empty
        -1.0, # Empty
        -1.0  # Empty
    ]]).T

    assert(rec.data.shape == (5, 1))
    assert_allclose(rec.data, expected, atol=1e-7)

    df = rec.to_dataframe()
    assert df.shape == (5, 1)
    assert_allclose(df.values, expected, atol=1e-7)



def test_numpy_array_level_recorder(simple_storage_model):
    model = simple_storage_model

    storage = model.nodes["Storage"]
    level_param = InterpolatedVolumeParameter(model, storage, [0, 20], [0, 100])
    storage.level = level_param
    level_rec = NumpyArrayLevelRecorder(model, storage, temporal_agg_func='min')

    model.run()

    expected = np.array([[50, 35, 20, 5, 0]]).T
    assert_allclose(level_rec.data, expected, atol=1e-7)

    df = level_rec.to_dataframe()
    assert df.shape == (5, 1)
    assert_allclose(df.values, expected, atol=1e-7)

    assert_allclose(level_rec.aggregated_value(), np.min(expected))


def test_numpy_array_area_recorder(simple_storage_model):

    model = simple_storage_model

    storage = model.nodes["Storage"]
    area_param = InterpolatedVolumeParameter(model, storage, [0, 20], [0, 100])
    storage.area = area_param
    area_rec = NumpyArrayAreaRecorder(model, storage, temporal_agg_func='min')

    model.run()

    expected = np.array([[50, 35, 20, 5, 0]]).T
    assert_allclose(area_rec.data, expected, atol=1e-7)

    df = area_rec.to_dataframe()
    assert df.shape == (5, 1)
    assert_allclose(df.values, expected, atol=1e-7)

    assert_allclose(area_rec.aggregated_value(), np.min(expected))


def test_numpy_parameter_recorder(simple_linear_model):
    """
    Test the NumpyArrayParameterRecorder
    """
    from pywr.parameters import DailyProfileParameter

    model = simple_linear_model
    # using leap year simplifies tests
    model.timestepper.start = pandas.to_datetime("2016-01-01")
    model.timestepper.end = pandas.to_datetime("2016-12-31")
    otpt = model.nodes['Output']

    p = DailyProfileParameter(model, np.arange(366, dtype=np.float64), )
    p.name = 'daily profile'
    model.nodes['Input'].max_flow = p
    otpt.cost = -2.0
    rec = NumpyArrayParameterRecorder(model, model.nodes['Input'].max_flow)

    # test retrieval of recorder
    assert model.recorders['numpyarrayparameterrecorder.daily profile'] == rec

    model.run()

    assert rec.data.shape == (366, 1)
    assert_allclose(rec.data, np.arange(366, dtype=np.float64)[:, np.newaxis])

    df = rec.to_dataframe()
    assert df.shape == (366, 1)
    assert_allclose(df.values, np.arange(366, dtype=np.float64)[:, np.newaxis])


def test_numpy_daily_profile_parameter_recorder(simple_linear_model):
    """
    Test the NumpyArrayDailyProfileParameterRecorder
    """
    from pywr.parameters import DailyProfileParameter

    model = simple_linear_model
    # using leap year simplifies tests
    model.timestepper.start = pandas.to_datetime("2016-01-01")
    model.timestepper.end = pandas.to_datetime("2017-12-31")
    otpt = model.nodes['Output']

    p = DailyProfileParameter(model, np.arange(366, dtype=np.float64), )
    p.name = 'daily profile'
    model.nodes['Input'].max_flow = p
    otpt.cost = -2.0
    rec = NumpyArrayDailyProfileParameterRecorder(model, model.nodes['Input'].max_flow)

    # test retrieval of recorder
    assert model.recorders['numpyarraydailyprofileparameterrecorder.daily profile'] == rec

    model.run()

    assert rec.data.shape == (366, 1)
    assert_allclose(rec.data, np.arange(366, dtype=np.float64)[:, np.newaxis])

    df = rec.to_dataframe()
    assert df.shape == (366, 1)
    assert_allclose(df.values, np.arange(366, dtype=np.float64)[:, np.newaxis])


def test_numpy_index_parameter_recorder(simple_storage_model):
    """
    Test the NumpyArrayIndexParameterRecorder

    Note the parameter is recorded at the start of the timestep, while the
    storage is recorded at the end of the timestep.
    """
    from pywr.parameters.control_curves import ControlCurveIndexParameter

    model = simple_storage_model

    res = model.nodes['Storage']

    p = ControlCurveIndexParameter(model, res, [5.0/20.0, 2.5/20.0])

    res_rec = NumpyArrayStorageRecorder(model, res)
    lvl_rec = NumpyArrayIndexParameterRecorder(model, p)

    model.run()

    assert(res_rec.data.shape == (5, 1))
    assert_allclose(res_rec.data, np.array([[7, 4, 1, 0, 0]]).T, atol=1e-7)
    assert (lvl_rec.data.shape == (5, 1))
    assert_allclose(lvl_rec.data, np.array([[0, 0, 1, 2, 2]]).T, atol=1e-7)


    df = lvl_rec.to_dataframe()
    assert df.shape == (5, 1)
    assert_allclose(df.values, np.array([[0, 0, 1, 2, 2]]).T, atol=1e-7)


def test_parameter_recorder_json():
    model = load_model("parameter_recorder.json")
    rec_demand = model.recorders["demand_max_recorder"]
    rec_supply = model.recorders["supply_max_recorder"]
    model.run()
    assert_allclose(rec_demand.data, 10)
    assert_allclose(rec_supply.data, 15)


def test_nested_recorder_json():
    model = load_model("agg_recorder_nesting.json")
    rec_demand = model.recorders["demand_max_recorder"]
    rec_supply = model.recorders["supply_max_recorder"]
    rec_total = model.recorders["max_recorder"]
    model.run()
    assert_allclose(rec_demand.aggregated_value(), 10)
    assert_allclose(rec_supply.aggregated_value(), 15)
    assert_allclose(rec_total.aggregated_value(), 25)


@pytest.fixture()
def daily_profile_model(simple_linear_model):
    model = simple_linear_model
    # using leap year simplifies test
    model.timestepper.start = pandas.to_datetime("2016-01-01")
    model.timestepper.end = pandas.to_datetime("2016-12-31")

    node = model.nodes["Input"]
    values = np.arange(0, 366, dtype=np.float64)
    node.max_flow = DailyProfileParameter(model, values, name='profile')
    return model


def test_parameter_mean_recorder(daily_profile_model):
    model = daily_profile_model
    node = model.nodes["Input"]
    scenario = Scenario(model, "dummy", size=3)

    timesteps = 3
    rec_mean = RollingWindowParameterRecorder(model, node.max_flow, timesteps,
                                              temporal_agg_func="mean", name="rec_mean")
    rec_sum = RollingWindowParameterRecorder(model, node.max_flow, timesteps,
                                             temporal_agg_func="sum", name="rec_sum")
    rec_min = RollingWindowParameterRecorder(model, node.max_flow, timesteps,
                                             temporal_agg_func="min", name="rec_min")
    rec_max = RollingWindowParameterRecorder(model, node.max_flow, timesteps,
                                             temporal_agg_func="max", name="rec_max")

    model.run()

    assert_allclose(rec_mean.data[[0, 1, 2, 3, 364], 0], [0, 0.5, 1, 2, 363])
    assert_allclose(rec_max.data[[0, 1, 2, 3, 364], 0], [0, 1, 2, 3, 364])
    assert_allclose(rec_min.data[[0, 1, 2, 3, 364], 0], [0, 0, 0, 1, 362])
    assert_allclose(rec_sum.data[[0, 1, 2, 3, 364], 0], [0, 1, 3, 6, 1089])

def test_parameter_mean_recorder_json(simple_linear_model):
    model = simple_linear_model
    node = model.nodes["Input"]
    values = np.arange(0, 366, dtype=np.float64)
    parameter = DailyProfileParameter(model, values, name="input_max_flow")

    node.max_flow = parameter

    data = {
        "type": "rollingwindowparameter",
        "parameter": "input_max_flow",
        "window": 3,
        "temporal_agg_func": "mean",
    }

    rec = load_recorder(model, data)


class TestTotalParameterRecorder:

    @pytest.mark.parametrize('factor, integrate',
                             [[1.0, False], [1.0, True], [2.0, False], [0.5, True]])
    def test_values(self, daily_profile_model, factor, integrate):
        model = daily_profile_model
        model.timestepper.delta = 2
        param = model.parameters['profile']
        rec = TotalParameterRecorder(model, param, name="total", factor=factor, integrate=integrate)
        model.run()

        expected = np.arange(0, 366, dtype=np.float64)[::2].sum()*factor
        if integrate:
            expected *= 2
        assert_allclose(rec.values(), expected)

    @pytest.mark.parametrize('integrate', [None, True, False])
    def test_from_json(self, daily_profile_model, integrate):

        model = daily_profile_model

        data = {
            "type": "totalparameter",
            "parameter": "profile",
            "agg_func": "mean",
            "factor": 2.0,
        }

        if integrate is not None:
            data['integrate'] = integrate

        rec = load_recorder(model, data)
        assert rec.factor == 2.0

        if integrate is not None:
            assert rec.integrate == integrate
        else:
            assert not rec.integrate


class TestMeanParameterRecorder:
    @pytest.mark.parametrize('factor', [1.0, 2.0, 0.5])
    def test_values(self, daily_profile_model, factor):
        model = daily_profile_model
        model.timestepper.delta = 2
        param = model.parameters['profile']
        rec = MeanParameterRecorder(model, param, name="mean", factor=factor)
        model.run()

        expected = np.arange(0, 366, dtype=np.float64)[::2].mean()*factor
        assert_allclose(rec.values(), expected)

    def test_from_json(self, daily_profile_model):

        model = daily_profile_model

        data = {
            "type": "meanparameter",
            "parameter": "profile",
            "agg_func": "mean",
            "factor": 2.0,
        }

        rec = load_recorder(model, data)
        assert rec.factor == 2.0


def test_concatenated_dataframes(simple_storage_model):
    """
    Test that Model.to_dataframe returns something sensible.

    """
    model = simple_storage_model

    scA = Scenario(model, 'A', size=2)
    scB = Scenario(model, 'B', size=3)

    res = model.nodes['Storage']
    rec1 = NumpyArrayStorageRecorder(model, res)
    otpt = model.nodes['Output']
    rec2 = NumpyArrayNodeRecorder(model, otpt)
    # The following can't return a DataFrame; is included to check
    # it doesn't cause any issues
    rec3 = TotalDeficitNodeRecorder(model, otpt)

    model.run()

    df = model.to_dataframe()
    assert df.shape == (5, 2*2*3)
    assert df.columns.names == ['Recorder', 'A', 'B']


@pytest.mark.parametrize("complib", [None, "gzip", "bz2"])
def test_csv_recorder(simple_storage_model, tmpdir, complib):
    """Test the CSV Recorder
    """
    model = simple_storage_model
    otpt = model.nodes['Output']
    otpt.cost = -2.0

    # Rename output to a unicode character to check encoding to files
    otpt.name = u"\u03A9"
    expected_header = ['Datetime', 'Input', 'Storage', u"\u03A9"]

    csvfile = tmpdir.join('output.csv')
    # By default the CSVRecorder saves all nodes in alphabetical order
    # and scenario index 0.
    rec = CSVRecorder(model, str(csvfile), complib=complib, complevel=5)

    model.run()

    import csv
    kwargs = {"encoding": "utf-8"}
    mode = "rt"

    if complib == "gzip":
        import gzip
        fh = gzip.open(str(csvfile), mode, **kwargs)
    elif complib in ("bz2", "bzip2"):
        import bz2
        fh = bz2.open(str(csvfile), mode, **kwargs)
    else:
        fh = open(str(csvfile), mode, **kwargs)

    expected = [
        expected_header,
        ['2016-01-01T00:00:00', 5.0, 7.0, 8.0],
        ['2016-01-02T00:00:00', 5.0, 4.0, 8.0],
        ['2016-01-03T00:00:00', 5.0, 1.0, 8.0],
        ['2016-01-04T00:00:00', 5.0, 0.0, 6.0],
        ['2016-01-05T00:00:00', 5.0, 0.0, 5.0],
    ]

    data = fh.read(1024)
    dialect = csv.Sniffer().sniff(data)
    fh.seek(0)
    reader = csv.reader(fh, dialect)

    for irow, row in enumerate(reader):
        expected_row = expected[irow]
        if irow == 0:
            assert expected_row == row
        else:
            assert expected_row[0] == row[0]  # Check datetime
            # Check values
            np.testing.assert_allclose([float(v) for v in row[1:]], expected_row[1:])
    fh.close()


def test_loading_csv_recorder_from_json(tmpdir):
    """
    Test the CSV Recorder which is loaded from json
    """

    filename = 'csv_recorder.json'

    # This is a bit horrible, but need to edit the JSON dynamically
    # so that the output.h5 is written in the temporary directory
    path = os.path.join(os.path.dirname(__file__), 'models')
    with open(os.path.join(path, filename), 'r') as f:
        data = f.read()
    data = json.loads(data)

    # Make an absolute, but temporary, path for the recorder
    url = data['recorders']['model_out']['url']
    data['recorders']['model_out']['url'] = str(tmpdir.join(url))

    model = Model.load(data, path=path)

    csvfile = tmpdir.join('output.csv')
    model.run()

    periods = model.timestepper.datetime_index

    import csv
    with open(str(csvfile), 'r') as fh:
        dialect = csv.Sniffer().sniff(fh.read(1024))
        fh.seek(0)
        reader = csv.reader(fh, dialect)
        for irow, row in enumerate(reader):
            if irow == 0:
                expected = ['Datetime', 'inpt', 'otpt']
                actual = row
            else:
                dt = periods[irow-1].to_timestamp()
                expected = [dt.isoformat()]
                actual = [row[0]]
                assert np.all((np.array([float(v) for v in row[1:]]) - 10.0) < 1e-12)
            assert expected == actual

class TestTablesRecorder:

    def test_create_directory(self, simple_linear_model, tmpdir):
        """ Test TablesRecorder to create a new directory """

        model = simple_linear_model
        otpt = model.nodes['Output']
        inpt = model.nodes['Input']
        agg_node = AggregatedNode(model, 'Sum', [otpt, inpt])

        inpt.max_flow = 10.0
        otpt.cost = -2.0
        # Make a path with a new directory
        folder = tmpdir.join('outputs')
        h5file = folder.join('output.h5')
        assert(not folder.exists())
        rec = TablesRecorder(model, str(h5file), create_directories=True)
        model.run()
        assert(folder.exists())
        assert(h5file.exists())

    def test_nodes(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder

        """
        model = simple_linear_model
        otpt = model.nodes['Output']
        inpt = model.nodes['Input']
        agg_node = AggregatedNode(model, 'Sum', [otpt, inpt])

        inpt.max_flow = 10.0
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            rec = TablesRecorder(model, h5f)

            model.run()

            for node_name in model.nodes.keys():
                ca = h5f.get_node('/', node_name)
                assert ca.shape == (365, 1)
                if node_name == 'Sum':
                    np.testing.assert_allclose(ca, 20.0)
                else:
                    np.testing.assert_allclose(ca, 10.0)

            from datetime import date, timedelta
            d = date(2015, 1, 1)
            time = h5f.get_node('/time')
            for i in range(len(model.timestepper)):
                row = time[i]
                assert row['year'] == d.year
                assert row['month'] == d.month
                assert row['day'] == d.day

                d += timedelta(1)

            scenarios = h5f.get_node('/scenarios')
            for i, s in enumerate(model.scenarios.scenarios):
                row = scenarios[i]
                assert row['name'] == s.name.encode('utf-8')
                assert row['size'] == s.size

            model.reset()
            model.run()

            time = h5f.get_node('/time')
            assert len(time) == len(model.timestepper)

    def test_multiple_scenarios(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder

        """
        from pywr.parameters import ConstantScenarioParameter
        model = simple_linear_model
        scA = Scenario(model, name='A', size=4)
        scB = Scenario(model, name='B', size=2)

        otpt = model.nodes['Output']
        inpt = model.nodes['Input']

        inpt.max_flow = ConstantScenarioParameter(model, scA, [10, 20, 30, 40])
        otpt.max_flow = ConstantScenarioParameter(model, scB, [20, 40])
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            rec = TablesRecorder(model, h5f)

            model.run()

            for node_name in model.nodes.keys():
                ca = h5f.get_node('/', node_name)
                assert ca.shape == (365, 4, 2)
                np.testing.assert_allclose(ca[0, ...], [[10, 10], [20, 20], [20, 30], [20, 40]])

            scenarios = h5f.get_node('/scenarios')
            for i, s in enumerate(model.scenarios.scenarios):
                row = scenarios[i]
                assert row['name'] == s.name.encode('utf-8')
                assert row['size'] == s.size

    def test_user_scenarios(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder with user defined scenario subset

        """
        from pywr.parameters import ConstantScenarioParameter
        model = simple_linear_model
        scA = Scenario(model, name='A', size=4)
        scB = Scenario(model, name='B', size=2)

        # Use first and last combinations
        model.scenarios.user_combinations = [[0, 0], [3, 1]]

        otpt = model.nodes['Output']
        inpt = model.nodes['Input']

        inpt.max_flow = ConstantScenarioParameter(model, scA, [10, 20, 30, 40])
        otpt.max_flow = ConstantScenarioParameter(model, scB, [20, 40])
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            rec = TablesRecorder(model, h5f)

            model.run()

            for node_name in model.nodes.keys():
                ca = h5f.get_node('/', node_name)
                assert ca.shape == (365, 2)
                np.testing.assert_allclose(ca[0, ...], [10, 40])

            # check combinations table exists
            combinations = h5f.get_node('/scenario_combinations')
            for i, comb in enumerate(model.scenarios.user_combinations):
                row = combinations[i]
                assert row['A'] == comb[0]
                assert row['B'] == comb[1]

    def test_parameters(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder

        """
        from pywr.parameters import ConstantParameter

        model = simple_linear_model
        otpt = model.nodes['Output']
        inpt = model.nodes['Input']

        p = ConstantParameter(model, 10.0, name='max_flow')
        inpt.max_flow = p

        # ensure TablesRecorder can handle parameters with a / in the name
        p_slash = ConstantParameter(model, 0.0, name='name with a / in it')
        inpt.min_flow = p_slash

        agg_node = AggregatedNode(model, 'Sum', [otpt, inpt])

        inpt.max_flow = 10.0
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            with pytest.warns(ParameterNameWarning):
                rec = TablesRecorder(model, h5f, parameters=[p, p_slash])

            # check parameters have been added to the component tree
            # this is particularly important for parameters which update their
            # values in `after`, e.g. DeficitParameter (see #465)
            assert(not model.find_orphaned_parameters())
            assert(p in rec.children)
            assert(p_slash in rec.children)

            with pytest.warns(tables.NaturalNameWarning):
                model.run()

            for node_name in model.nodes.keys():
                ca = h5f.get_node('/', node_name)
                assert ca.shape == (365, 1)
                if node_name == 'Sum':
                    np.testing.assert_allclose(ca, 20.0)
                elif "name with a" in node_name:
                    assert(node_name == "name with a _ in it")
                    np.testing.assert_allclose(ca, 0.0)
                else:
                    np.testing.assert_allclose(ca, 10.0)

    def test_nodes_with_str(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder

        """
        from pywr.parameters import ConstantParameter

        model = simple_linear_model
        otpt = model.nodes['Output']
        inpt = model.nodes['Input']
        agg_node = AggregatedNode(model, 'Sum', [otpt, inpt])
        p = ConstantParameter(model, 10.0, name='max_flow')
        inpt.max_flow = p

        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            nodes = ['Output', 'Input', 'Sum']
            where = "/agroup"
            rec = TablesRecorder(model, h5f, nodes=nodes,
                                 parameters=[p, ], where=where)

            model.run()

            for node_name in ['Output', 'Input', 'Sum', 'max_flow']:
                ca = h5f.get_node("/agroup/" + node_name)
                assert ca.shape == (365, 1)
                if node_name == 'Sum':
                    np.testing.assert_allclose(ca, 20.0)
                else:
                    np.testing.assert_allclose(ca, 10.0)

    def test_demand_saving_with_indexed_array(self, tmpdir):
        """Test recording various items from demand saving example

        """
        model = load_model("demand_saving2.json")

        model.timestepper.end = "2016-01-31"

        model.check()

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:

            nodes = [
                ('/outputs/demand', 'Demand'),
                ('/storage/reservoir', 'Reservoir'),
            ]

            parameters = [
                ('/parameters/demand_saving_level', 'demand_saving_level'),
            ]

            rec = TablesRecorder(model, h5f, nodes=nodes, parameters=parameters)

            model.run()

            max_volume = model.nodes["Reservoir"].max_volume
            rec_demand = h5f.get_node('/outputs/demand', 'Demand').read()
            rec_storage = h5f.get_node('/storage/reservoir', 'Reservoir').read()

            # model starts with no demand saving
            demand_baseline = 50.0
            demand_factor = 0.9  # jan-apr
            demand_saving = 1.0
            assert_allclose(rec_demand[0, 0], demand_baseline * demand_factor * demand_saving)

            # first control curve breached
            demand_saving = 0.95
            assert (rec_storage[4, 0] < (0.8 * max_volume))
            assert_allclose(rec_demand[5, 0], demand_baseline * demand_factor * demand_saving)

            # second control curve breached
            demand_saving = 0.5
            assert (rec_storage[11, 0] < (0.5 * max_volume))
            assert_allclose(rec_demand[12, 0], demand_baseline * demand_factor * demand_saving)

    def test_demand_saving_with_indexed_array(self, tmpdir):
        """Test recording various items from demand saving example.

        This time the TablesRecorder is defined in JSON.
        """
        filename = "demand_saving_with_tables_recorder.json"
        # This is a bit horrible, but need to edit the JSON dynamically
        # so that the output.h5 is written in the temporary directory
        path = os.path.join(os.path.dirname(__file__), 'models')
        with open(os.path.join(path, filename), 'r') as f:
            data = f.read()
        data = json.loads(data)

        # Make an absolute, but temporary, path for the recorder
        url = data['recorders']['database']['url']
        data['recorders']['database']['url'] = str(tmpdir.join(url))

        model = Model.load(data, path=path)

        model.timestepper.end = "2016-01-31"
        model.check()

        # run model
        model.run()

        # run model again (to test reset behaviour)
        model.run()
        max_volume = model.nodes["Reservoir"].max_volume

        h5file = tmpdir.join('output.h5')
        with tables.open_file(str(h5file), 'r') as h5f:
            assert model.metadata['title'] == h5f.title
            # Check metadata on root node
            assert h5f.root._v_attrs.author == 'pytest'
            assert h5f.root._v_attrs.run_number == 0

            rec_demand = h5f.get_node('/outputs/demand').read()
            rec_storage = h5f.get_node('/storage/reservoir').read()

            # model starts with no demand saving
            demand_baseline = 50.0
            demand_factor = 0.9  # jan-apr
            demand_saving = 1.0
            assert_allclose(rec_demand[0, 0], demand_baseline * demand_factor * demand_saving)

            # first control curve breached
            demand_saving = 0.95
            assert (rec_storage[4, 0] < (0.8 * max_volume))
            assert_allclose(rec_demand[5, 0], demand_baseline * demand_factor * demand_saving)

            # second control curve breached
            demand_saving = 0.5
            assert (rec_storage[11, 0] < (0.5 * max_volume))
            assert_allclose(rec_demand[12, 0], demand_baseline * demand_factor * demand_saving)

    @pytest.mark.skipif(Model().solver.name == "glpk-edge", reason="Not valid for GLPK Edge based solver.")
    def test_routes(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder

        """
        model = simple_linear_model
        otpt = model.nodes['Output']
        inpt = model.nodes['Input']
        agg_node = AggregatedNode(model, 'Sum', [otpt, inpt])

        inpt.max_flow = 10.0
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            rec = TablesRecorder(model, h5f, routes_flows='flows')

            model.run()

            flows = h5f.get_node('/flows')
            assert flows.shape == (365, 1, 1)
            np.testing.assert_allclose(flows.read(), np.ones((365, 1, 1))*10)

            routes = h5f.get_node('/routes')
            assert routes.shape[0] == 1
            row = routes[0]
            row['start'] = "Input"
            row['end'] = "Output"

            from datetime import date, timedelta
            d = date(2015, 1, 1)
            time = h5f.get_node('/time')
            for i in range(len(model.timestepper)):
                row = time[i]
                assert row['year'] == d.year
                assert row['month'] == d.month
                assert row['day'] == d.day

                d += timedelta(1)

            scenarios = h5f.get_node('/scenarios')
            for s in model.scenarios.scenarios:
                row = scenarios[i]
                assert row['name'] == s.name
                assert row['size'] == s.size

            model.reset()
            model.run()

            time = h5f.get_node('/time')
            assert len(time) == len(model.timestepper)

    @pytest.mark.skipif(Model().solver.name == "glpk-edge", reason="Not valid for GLPK Edge based solver.")
    def test_routes_multiple_scenarios(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder

        """
        from pywr.parameters import ConstantScenarioParameter
        model = simple_linear_model
        scA = Scenario(model, name='A', size=4)
        scB = Scenario(model, name='B', size=2)

        otpt = model.nodes['Output']
        inpt = model.nodes['Input']

        inpt.max_flow = ConstantScenarioParameter(model, scA, [10, 20, 30, 40])
        otpt.max_flow = ConstantScenarioParameter(model, scB, [20, 40])
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            rec = TablesRecorder(model, h5f, routes_flows='flows')

            model.run()

            flows = h5f.get_node('/flows')
            assert flows.shape == (365, 1, 4, 2)
            np.testing.assert_allclose(flows[0, 0], [[10, 10], [20, 20], [20, 30], [20, 40]])

    @pytest.mark.skipif(Model().solver.name == "glpk-edge", reason="Not valid for GLPK Edge based solver.")
    def test_routes_user_scenarios(self, simple_linear_model, tmpdir):
        """
        Test the TablesRecorder with user defined scenario subset

        """
        from pywr.parameters import ConstantScenarioParameter
        model = simple_linear_model
        scA = Scenario(model, name='A', size=4)
        scB = Scenario(model, name='B', size=2)

        # Use first and last combinations
        model.scenarios.user_combinations = [[0, 0], [3, 1]]

        otpt = model.nodes['Output']
        inpt = model.nodes['Input']

        inpt.max_flow = ConstantScenarioParameter(model, scA, [10, 20, 30, 40])
        otpt.max_flow = ConstantScenarioParameter(model, scB, [20, 40])
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        import tables
        with tables.open_file(str(h5file), 'w') as h5f:
            rec = TablesRecorder(model, h5f, routes_flows='flows')

            model.run()

            flows = h5f.get_node('/flows')
            assert flows.shape == (365, 1, 2)
            np.testing.assert_allclose(flows[0, 0], [10, 40])

            # check combinations table exists
            combinations = h5f.get_node('/scenario_combinations')
            for i, comb in enumerate(model.scenarios.user_combinations):
                row = combinations[i]
                assert row['A'] == comb[0]
                assert row['B'] == comb[1]

        # This part of the test requires IPython (see `pywr.notebook`)
        pytest.importorskip("IPython")  # triggers a skip of the test if IPython not found.
        from pywr.notebook.sankey import routes_to_sankey_links

        links = routes_to_sankey_links(str(h5file), 'flows')
        # Value is mean of 10 and 40

        link = links[0]
        assert link['source'] == 'Input'
        assert link['target'] == 'Output'
        np.testing.assert_allclose(link['value'], 25.0)

        links = routes_to_sankey_links(str(h5file), 'flows', scenario_slice=0)
        link = links[0]
        assert link['source'] == 'Input'
        assert link['target'] == 'Output'
        np.testing.assert_allclose(link['value'], 10.0)

        links = routes_to_sankey_links(str(h5file), 'flows', scenario_slice=1, time_slice=0)
        link = links[0]
        assert link['source'] == 'Input'
        assert link['target'] == 'Output'
        np.testing.assert_allclose(link['value'], 40.0)

    def test_generate_dataframes(self, simple_linear_model, tmpdir):
        """Test TablesRecorder.generate_dataframes """
        from pywr.parameters import ConstantScenarioParameter
        model = simple_linear_model
        scA = Scenario(model, name='A', size=4)
        scB = Scenario(model, name='B', size=2)

        otpt = model.nodes['Output']
        inpt = model.nodes['Input']

        inpt.max_flow = ConstantScenarioParameter(model, scA, [10, 20, 30, 40])
        otpt.max_flow = ConstantScenarioParameter(model, scB, [20, 40])
        otpt.cost = -2.0

        h5file = tmpdir.join('output.h5')
        TablesRecorder(model, h5file)
        model.run()

        dfs = {}
        for node, df in TablesRecorder.generate_dataframes(h5file):
            dfs[node] = df

        for node_name in model.nodes.keys():
            df = dfs[node_name]
            assert df.shape == (365, 8)
            np.testing.assert_allclose(df.iloc[0, :], [10, 10, 20, 20, 20, 30, 20, 40])


class TestDeficitRecorders:

    @pytest.mark.parametrize('demand', [30.0, 0.0])
    def test_total_deficit_node_recorder(self, simple_linear_model, demand):
        """
        Test TotalDeficitNodeRecorder
        """
        model = simple_linear_model
        model.timestepper.delta = 5
        otpt = model.nodes['Output']
        otpt.max_flow = demand
        model.nodes['Input'].max_flow = 10.0
        otpt.cost = -2.0
        rec = TotalDeficitNodeRecorder(model, otpt)

        model.step()
        assert_allclose(max(demand - 10.0, 0.0)*5, rec.aggregated_value(), atol=1e-7)

        model.step()
        assert_allclose(2*max(demand - 10.0, 0.0)*5, rec.aggregated_value(), atol=1e-7)

    def test_array_deficit_recoder(self, simple_linear_model):
        """Test `NumpyArrayNodeDeficitRecorder` """
        model = simple_linear_model
        model.timestepper.delta = 1
        otpt = model.nodes['Output']

        inflow = np.arange(365) * 0.1
        demand = np.ones_like(inflow) * 30.0

        model.nodes['Input'].max_flow = ArrayIndexedParameter(model, inflow)
        otpt.max_flow = ArrayIndexedParameter(model, demand)
        otpt.cost = -2.0

        expected_supply = np.minimum(inflow, demand)
        expected_deficit = demand - expected_supply

        rec = NumpyArrayNodeDeficitRecorder(model, otpt)

        model.run()

        assert rec.data.shape == (365, 1)
        np.testing.assert_allclose(expected_deficit[:, np.newaxis], rec.data)

        df = rec.to_dataframe()
        assert df.shape == (365, 1)
        np.testing.assert_allclose(expected_deficit[:, np.newaxis], df.values)

    @pytest.mark.parametrize('demand_factor', [30.0, 0.0])
    def test_array_supplied_ratio_recoder(self, simple_linear_model, demand_factor):
        """Test `NumpyArrayNodeSuppliedRatioRecorder` """
        model = simple_linear_model
        model.timestepper.delta = 1
        otpt = model.nodes['Output']

        inflow = np.arange(365) * 0.1
        demand = np.ones_like(inflow) * demand_factor

        model.nodes['Input'].max_flow = ArrayIndexedParameter(model, inflow)
        otpt.max_flow = ArrayIndexedParameter(model, demand)
        otpt.cost = -2.0

        expected_supply = np.minimum(inflow, demand)
        if demand_factor == 0.0:
            expected_ratio = np.ones_like(expected_supply)
        else:
            expected_ratio = expected_supply / demand

        rec = NumpyArrayNodeSuppliedRatioRecorder(model, otpt)

        model.run()

        assert rec.data.shape == (365, 1)
        np.testing.assert_allclose(expected_ratio[:, np.newaxis], rec.data)

        df = rec.to_dataframe()
        assert df.shape == (365, 1)
        np.testing.assert_allclose(expected_ratio[:, np.newaxis], df.values)

    @pytest.mark.parametrize('demand_factor', [30.0, 0.0])
    def test_array_curtailment_ratio_recoder(self, simple_linear_model, demand_factor):
        """Test `NumpyArrayNodeCurtailmentRatioRecorder` """
        model = simple_linear_model
        model.timestepper.delta = 1
        otpt = model.nodes['Output']

        inflow = np.arange(365) * 0.1
        demand = np.ones_like(inflow) * demand_factor

        model.nodes['Input'].max_flow = ArrayIndexedParameter(model, inflow)
        otpt.max_flow = ArrayIndexedParameter(model, demand)
        otpt.cost = -2.0

        expected_supply = np.minimum(inflow, demand)
        if demand_factor == 0.0:
            expected_curtailment_ratio = np.zeros_like(expected_supply)
        else:
            expected_curtailment_ratio = 1 - expected_supply / demand

        rec = NumpyArrayNodeCurtailmentRatioRecorder(model, otpt)

        model.run()

        assert rec.data.shape == (365, 1)
        np.testing.assert_allclose(expected_curtailment_ratio[:, np.newaxis], rec.data)

        df = rec.to_dataframe()
        assert df.shape == (365, 1)
        np.testing.assert_allclose(expected_curtailment_ratio[:, np.newaxis], df.values)


def test_timestep_count_index_parameter_recorder(simple_storage_model):
    """
    The test uses a simple reservoir model with different inputs that
    trigger a control curve failure after a different number of years.
    """
    from pywr.parameters import ConstantScenarioParameter, ConstantParameter
    from pywr.parameters.control_curves import ControlCurveIndexParameter
    model = simple_storage_model
    scenario = Scenario(model, 'A', size=2)
    # Simulate 5 years
    model.timestepper.start = '2015-01-01'
    model.timestepper.end = '2019-12-31'
    # Control curve parameter
    param = ControlCurveIndexParameter(model, model.nodes['Storage'], ConstantParameter(model, 0.25))

    # Storage model has a capacity of 20, but starts at 10 Ml
    # Demand is roughly 2 Ml/d per year
    #  First ensemble balances the demand
    #  Second ensemble should fail during 3rd year
    demand = 2.0 / 365
    model.nodes['Input'].max_flow = ConstantScenarioParameter(model, scenario, [demand, 0])
    model.nodes['Output'].max_flow = demand

    # Create the recorder with a threshold of 1
    rec = TimestepCountIndexParameterRecorder(model, param, 1)

    model.run()

    assert_allclose([0, 183 + 365 + 365], rec.values(), atol=1e-7)


@pytest.mark.parametrize(("params", "exclude_months"), [
    [1, None],
    [2, None],
    [1, [1, 2, 12]],
])
def test_annual_count_index_threshold_recorder(simple_storage_model, params, exclude_months):
    """
    The test sets uses a simple reservoir model with different inputs that
    trigger a control curve failure after different numbers of years.
    """
    from pywr.parameters import ConstantScenarioParameter, ConstantParameter
    from pywr.parameters.control_curves import ControlCurveIndexParameter
    model = simple_storage_model
    scenario = Scenario(model, 'A', size=2)
    # Simulate 5 years
    model.timestepper.start = '2015-01-01'
    model.timestepper.end = '2019-12-31'
    # Control curve parameter
    param = ControlCurveIndexParameter(model, model.nodes['Storage'], ConstantParameter(model, 0.25))

    # Storage model has a capacity of 20, but starts at 10 Ml
    # Demand is roughly 2 Ml/d per year
    #  First ensemble balances the demand
    #  Second ensemble should fail during 3rd year
    demand = 2.0 / 365
    model.nodes['Input'].max_flow = ConstantScenarioParameter(model, scenario, [demand, 0])
    model.nodes['Output'].max_flow = demand

    # Create the recorder with a threshold of 1
    rec = AnnualCountIndexThresholdRecorder(model, [param] * params, 'TestRec', 1, exclude_months=exclude_months)

    model.run()

    # We expect no failures in the first ensemble, the reservoir starts failing halfway through
    # the 3rd year
    if exclude_months is None:
        expected_data = [[0, 0], [0, 0], [0, 183], [0, 365],[0, 365]]
    else:
        # Ignore counts for Jan, Feb and Dec
        assert exclude_months == [1, 2, 12]  # Test is hard-coded for these exclusion months.
        expected_data = [[0, 0], [0, 0], [0, 183 - 31], [0, 365 - 31 - 28 - 31], [0, 365 - 31 - 28 - 31]]

    assert_allclose(expected_data, rec.data, atol=1e-7)
    df = rec.to_dataframe()
    assert_allclose(expected_data, df.values, atol=1e-7)


class TestAnnualTotalFlowRecorder:

    def test_annual_total_flow_recorder(self, simple_linear_model):
        """
        Test AnnualTotalFlowRecorder
        """

        model = simple_linear_model
        otpt = model.nodes['Output']
        otpt.max_flow = 30.0
        model.nodes['Input'].max_flow = 10.0
        otpt.cost = -2
        rec = AnnualTotalFlowRecorder(model, 'Total Flow', [otpt])

        model.run()

        assert_allclose(3650.0, rec.data, atol=1e-7)
        df = rec.to_dataframe()
        assert_allclose([[3650.0]], df.values)

    def test_annual_total_flow_recorder_factored(self, simple_linear_model):
        """
        Test AnnualTotalFlowRecorder with factors applied
        """
        model = simple_linear_model
        otpt = model.nodes['Output']
        otpt.max_flow = 30.0
        inpt = model.nodes['Input']
        inpt.max_flow = 10.0
        otpt.cost = -2

        factors = [2.0, 1.0]
        rec_fact = AnnualTotalFlowRecorder(model, 'Total Flow', [otpt, inpt], factors=factors)

        model.run()

        assert_allclose(3650.0*3, rec_fact.data, atol=1e-7)
        df = rec_fact.to_dataframe()
        assert_allclose([[3650.0*3]], df.values)

    @pytest.mark.parametrize("end_date, expected", [("2013-01-04", [30.0, 40.0]), ("2012-12-31", [30.0]),])
    def test_annual_total_flow_recorder_year_end(self, simple_linear_model, end_date, expected):
        """
        Test AnnualTotalFlowRecorder when timestep crosses year end.

        The two parameterisations of this test run a single timestep of 7 days that crosses into the next year. In the
        first, the model end date is set to the end of the timestep, so flow for the recorder is assigned to each year
        according to the numbers of days of the timestep that are in each year (3 in the current, 4 in the next). In the
        second parameterisation, the model end date is set to the end of year, so flow for the recorder is only assigned
        for the first 3 days of the timestep that are in the current year. The subsequent 4 days of the timestep are not
        recorded because they are beyond the model end date.
        """
        model = simple_linear_model
        simple_linear_model.timestepper.start = "2012-12-29"
        simple_linear_model.timestepper.end = end_date
        simple_linear_model.timestepper.delta = "7D"

        otpt = model.nodes['Output']
        otpt.max_flow = 30.0
        model.nodes['Input'].max_flow = 10.0
        otpt.cost = -2
        rec = AnnualTotalFlowRecorder(model, 'Total Flow', [otpt])

        model.run()

        assert_allclose(expected, rec.data.flatten())


def test_total_flow_node_recorder(simple_linear_model):
    """
    Test TotalDeficitNodeRecorder
    """
    model = simple_linear_model
    otpt = model.nodes['Output']
    otpt.max_flow = 30.0
    model.nodes['Input'].max_flow = 10.0
    otpt.cost = -2.0
    rec = TotalFlowNodeRecorder(model, otpt)

    model.step()
    assert_allclose(10.0, rec.aggregated_value(), atol=1e-7)

    model.step()
    assert_allclose(20.0, rec.aggregated_value(), atol=1e-7)


def test_mean_flow_node_recorder(simple_linear_model):
    """
    Test MeanFlowNodeRecorder
    """
    model = simple_linear_model
    nt = len(model.timestepper)

    otpt = model.nodes['Output']
    otpt.max_flow = 30.0
    model.nodes['Input'].max_flow = 10.0
    otpt.cost = -2.0
    rec = MeanFlowNodeRecorder(model, otpt)

    model.run()
    assert_allclose(10.0, rec.aggregated_value(), atol=1e-7)


def custom_test_func(array, axis=None):
    return np.sum(array**2, axis=axis)


class TestAggregatedRecorder:
    """Tests for AggregatedRecorder"""
    funcs = {"min": np.min, "max": np.max, "mean": np.mean, "sum": np.sum}

    @pytest.mark.parametrize("agg_func", ["min", "max", "mean", "sum"])
    def test_aggregated_recorder(self, simple_linear_model, agg_func):
        model = simple_linear_model
        otpt = model.nodes['Output']
        otpt.max_flow = 30.0
        model.nodes['Input'].max_flow = 10.0
        otpt.cost = -2.0
        rec1 = TotalFlowNodeRecorder(model, otpt)
        rec2 = TotalDeficitNodeRecorder(model, otpt)

        func = TestAggregatedRecorder.funcs[agg_func]

        rec = AggregatedRecorder(model, [rec1, rec2], agg_func=agg_func)

        assert(rec in rec1.parents)
        assert(rec in rec2.parents)

        model.step()
        assert_allclose(func([10.0, 20.0]), rec.aggregated_value(), atol=1e-7)

        model.step()
        assert_allclose(func([20.0, 40.0]), rec.aggregated_value(), atol=1e-7)

    @pytest.mark.parametrize("agg_func", ["min", "max", "mean", "sum", "custom"])
    def test_agg_func_get_set(self, simple_linear_model, agg_func):
        """Test getter and setter for AggregatedRecorder.agg_func"""
        if agg_func == "custom":
            agg_func = custom_test_func
        model = simple_linear_model
        rec = AggregatedRecorder(model, [], agg_func=agg_func)
        assert rec.agg_func == agg_func
        rec.agg_func = "product"
        assert rec.agg_func == "product"


def test_reset_timestepper_recorder():
    model = Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-01')
    )

    inpt = Input(model, "input", max_flow=10)
    otpt = Output(model, "output", max_flow=50, cost=-10)
    inpt.connect(otpt)

    rec = NumpyArrayNodeRecorder(model, otpt)

    model.run()

    model.timestepper.end = pandas.to_datetime("2016-01-02")

    model.run()

def test_mean_flow_recorder():
    model = Model()
    model.timestepper.start = pandas.to_datetime("2016-01-01")
    model.timestepper.end = pandas.to_datetime("2016-01-04")

    inpt = Input(model, "input")
    otpt = Output(model, "output")
    inpt.connect(otpt)

    rec_flow = NumpyArrayNodeRecorder(model, inpt)
    rec_mean = RollingMeanFlowNodeRecorder(model, node=inpt, timesteps=3)

    scenario = Scenario(model, "dummy", size=2)

    inpt.max_flow = inpt.min_flow = FunctionParameter(model, inpt, lambda model, t, si: 2 + t.index)
    model.run()

    expected = [
        2.0,
        (2.0 + 3.0) / 2,
        (2.0 + 3.0 + 4.0) / 3,
        (3.0 + 4.0 + 5.0) / 3,  # zeroth day forgotten
    ]

    for value, expected_value in zip(rec_mean.data[:, 0], expected):
        assert_allclose(value, expected_value)

def test_mean_flow_recorder_days():
    model = Model()
    model.timestepper.delta = 7

    inpt = Input(model, "input")
    otpt = Output(model, "output")
    inpt.connect(otpt)

    rec_mean = RollingMeanFlowNodeRecorder(model, node=inpt, days=31)

    model.run()
    assert(rec_mean.timesteps == 4)

def test_mean_flow_recorder_json():
    model = load_model("mean_flow_recorder.json")

    # TODO: it's not possible to define a FunctionParameter in JSON yet
    supply1 = model.nodes["supply1"]
    supply1.max_flow = supply1.min_flow = FunctionParameter(model, supply1, lambda model, t, si: 2 + t.index)

    assert(len(model.recorders) == 3)

    rec_flow = model.recorders["Supply"]
    rec_mean = model.recorders["Mean Flow"]
    rec_check = model.recorders["Supply 2"]

    model.run()

    assert_allclose(rec_flow.data[:,0], [2.0, 3.0, 4.0, 5.0])
    assert_allclose(rec_mean.data[:,0], [2.0, 2.5, 3.0, 4.0])
    assert_allclose(rec_check.data[:,0], [50.0, 50.0, 60.0, 60.0])

def test_annual_count_index_parameter_recorder(simple_storage_model):
    """ Test AnnualCountIndexParameterRecord

    The test sets uses a simple reservoir model with different inputs that
    trigger a control curve failure after different numbers of years.
    """
    from pywr.parameters import ConstantScenarioParameter, ConstantParameter
    from pywr.parameters.control_curves import ControlCurveIndexParameter
    model = simple_storage_model
    scenario = Scenario(model, 'A', size=2)
    # Simulate 5 years
    model.timestepper.start = '2015-01-01'
    model.timestepper.end = '2019-12-31'
    # Control curve parameter
    param = ControlCurveIndexParameter(model, model.nodes['Storage'], ConstantParameter(model, 0.25))

    # Storage model has a capacity of 20, but starts at 10 Ml
    # Demand is roughly 2 Ml/d per year
    #  First ensemble balances the demand
    #  Second ensemble should fail during 3rd year
    demand = 2.0 / 365
    model.nodes['Input'].max_flow = ConstantScenarioParameter(model, scenario, [demand, 0])
    model.nodes['Output'].max_flow = demand

    # Create the recorder with a threshold of 1
    rec = AnnualCountIndexParameterRecorder(model, param, 1)

    model.run()
    # We expect no failures in the first ensemble, but 3 out of 5 in the second
    assert_allclose(rec.values(), [0, 3])


# The following fixtures are used for testing the recorders in
#  pywr.recorders.calibration which require an observed data set
#  to compare with the model prediction.

@pytest.fixture
def timeseries2_model():
    return load_model('timeseries2.json')


@pytest.fixture
def timeseries2_observed():
    path = os.path.join(os.path.dirname(__file__), 'models')
    df = pandas.read_csv(os.path.join(path, 'timeseries2.csv'),
                         parse_dates=True, dayfirst=True, index_col=0)
    df = df.asfreq(pandas.infer_freq(df.index))
    # perturb a bit
    df += np.random.normal(size=df.shape)
    return df

class TestCalibrationRecorders:
    data = [
        (RootMeanSquaredErrorNodeRecorder, lambda sim, obs: np.sqrt(np.mean((sim-obs)**2, axis=0))),
        (MeanAbsoluteErrorNodeRecorder, lambda sim, obs: np.mean(np.abs(sim-obs), axis=0)),
        (MeanSquareErrorNodeRecorder, lambda sim, obs: np.mean((sim-obs)**2, axis=0)),
        (PercentBiasNodeRecorder, lambda sim, obs: np.sum(obs-sim, axis=0)*100/np.sum(obs, axis=0)),
        (RMSEStandardDeviationRatioNodeRecorder, lambda sim, obs: np.sqrt(np.mean((obs-sim)**2, axis=0))/np.std(obs, axis=0)),
        (NashSutcliffeEfficiencyNodeRecorder, lambda sim, obs: 1.0 - np.sum((obs-sim)**2, axis=0)/np.sum((obs-obs.mean())**2, axis=0)),
    ]
    ids = ["rmse", "mae", "mse", "pbias", "rmse", "ns"]

    @pytest.mark.parametrize("cls,func", data, ids=ids)
    def test_calibration_recorder(self, timeseries2_model, timeseries2_observed, cls, func):
        model = timeseries2_model
        observed = timeseries2_observed
        node = model.nodes["river1"]
        recorder = cls(model, node, observed)

        model.run()

        simulated = model.nodes["catchment1"].max_flow.dataframe
        metric = func(simulated, observed)
        values = recorder.values()
        assert(values.shape[0] == len(model.scenarios.combinations))
        assert(values.ndim == 1)
        assert_allclose(metric, values)


@pytest.fixture
def cyclical_storage_model(simple_storage_model):
    """ Extends simple_storage_model to have a cyclical boundary condition """
    from pywr.parameters import AnnualHarmonicSeriesParameter, ConstantScenarioParameter
    m = simple_storage_model
    s = Scenario(m, name='Scenario A', size=3)

    m.timestepper.end = '2017-12-31'
    m.timestepper.delta = 5

    inpt = m.nodes['Input']
    inpt.max_flow = AnnualHarmonicSeriesParameter(m, 5, [0.1, 0.0, 0.25], [0.0, 0.0, 0.0], name="inpt_flow")

    otpt = m.nodes['Output']
    otpt.max_flow = ConstantScenarioParameter(m, s, [5, 6, 2])

    return m


@pytest.fixture
def cyclical_linear_model(simple_linear_model):
    """ Extends simple_storage_model to have a cyclical boundary condition """
    from pywr.parameters import AnnualHarmonicSeriesParameter, ConstantScenarioParameter
    m = simple_linear_model
    s = Scenario(m, name='Scenario A', size=3)

    m.timestepper.end = '2017-12-31'
    m.timestepper.delta = 5

    inpt = m.nodes['Input']
    inpt.max_flow = AnnualHarmonicSeriesParameter(m, 5, [1.0, 0.0, 0.5], [0.0, 0.0, 0.0])

    otpt = m.nodes['Output']
    otpt.max_flow = ConstantScenarioParameter(m, s, [5, 6, 2])
    otpt.cost = -10.0

    return m


class TestEventRecorder:
    """ Tests for EventRecorder """
    funcs = {"min": np.min, "max": np.max, "mean": np.mean, "median": np.median, "sum": np.sum}

    @pytest.mark.parametrize("threshold_component", [StorageThresholdRecorder, StorageThresholdParameter])
    def test_load(self, cyclical_storage_model, threshold_component):
        """Test load method"""
        m = cyclical_storage_model
        strg = m.nodes['Storage']
        param = threshold_component(m, strg, 4.0, predicate='<=', name="trigger")
        EventRecorder.load(m, {"name": "event_rec", "threshold": "trigger", "tracked_parameter": "inpt_flow"})
        EventDurationRecorder.load(m, {"event_recorder": "event_rec"})
        EventStatisticRecorder.load(m, {"event_recorder": "event_rec"})
        m.run()

    @pytest.mark.parametrize("recorder_agg_func", ["min", "max", "mean", "median", "sum"])
    def test_event_capture_with_storage(self, cyclical_storage_model, recorder_agg_func):
        """ Test Storage events using a StorageThresholdRecorder """
        m = cyclical_storage_model

        strg = m.nodes['Storage']
        arry = NumpyArrayStorageRecorder(m, strg)

        # Create the trigger using a threhsold parameter
        trigger = StorageThresholdRecorder(m, strg, 4.0, predicate='<=')
        evt_rec = EventRecorder(m, trigger)
        evt_dur = EventDurationRecorder(m, evt_rec, recorder_agg_func=recorder_agg_func, agg_func='max')

        m.run()

        # Ensure there is at least one event
        assert evt_rec.events

        # Build a timeseries of when the events say an event is active
        triggered = np.zeros_like(arry.data, dtype=np.int32)
        for evt in evt_rec.events:
            triggered[evt.start.index:evt.end.index, evt.scenario_index.global_id] = 1

            # Check the duration
            td = evt.end.datetime - evt.start.datetime
            assert evt.duration == td.days

        #   Test that the volumes in the Storage node during the event periods match
        assert_equal(triggered, arry.data <= 4)

        df = evt_rec.to_dataframe()

        assert len(df) == len(evt_rec.events)

        func = TestEventRecorder.funcs[recorder_agg_func]

        # Now check the EventDurationRecorder does the aggregation correctly
        expected_durations = []
        for si in m.scenarios.combinations:
            event_durations = []
            for evt in evt_rec.events:
                if evt.scenario_index.global_id == si.global_id:
                    event_durations.append(evt.duration)

            # If there are no events then the metric is zero
            if len(event_durations) > 0:
                expected_durations.append(func(event_durations))
            else:
                expected_durations.append(0.0)

        assert_allclose(evt_dur.values(), expected_durations)
        assert_allclose(evt_dur.aggregated_value(), np.max(expected_durations))

    def test_event_capture_with_node(self, cyclical_linear_model):
        """ Test Node flow events using a NodeThresholdRecorder """
        m = cyclical_linear_model

        otpt = m.nodes['Output']
        arry = NumpyArrayNodeRecorder(m, otpt)

        # Create the trigger using a threhsold parameter
        trigger = NodeThresholdRecorder(m, otpt, 4.0, predicate='>')
        evt_rec = EventRecorder(m, trigger)

        m.run()

        # Ensure there is at least one event
        assert evt_rec.events

        # Build a timeseries of when the events say an event is active
        triggered = np.zeros_like(arry.data, dtype=np.int32)
        for evt in evt_rec.events:
            triggered[evt.start.index:evt.end.index, evt.scenario_index.global_id] = 1

            # Check the duration
            td = evt.end.datetime - evt.start.datetime
            assert evt.duration == td.days

        # Test that the volumes in the Storage node during the event periods match
        assert_equal(triggered, arry.data > 4)

    @pytest.mark.parametrize("recorder_agg_func", ["min", "max", "mean", "median", "sum"])
    def test_no_event_capture_with_storage(self, cyclical_storage_model, recorder_agg_func):
        """ Test Storage events using a StorageThresholdRecorder """
        m = cyclical_storage_model

        strg = m.nodes['Storage']
        arry = NumpyArrayStorageRecorder(m, strg)

        # Create the trigger using a threhsold parameter
        trigger = StorageThresholdRecorder(m, strg, -1.0, predicate='<')
        evt_rec = EventRecorder(m, trigger)
        evt_dur = EventDurationRecorder(m, evt_rec, recorder_agg_func=recorder_agg_func, agg_func='max')

        m.run()

        # Ensure there are no events in this test
        assert len(evt_rec.events) == 0
        df = evt_rec.to_dataframe()
        assert len(df) == 0

        assert_allclose(evt_dur.values(), np.zeros(len(m.scenarios.combinations)))
        assert_allclose(evt_dur.aggregated_value(), 0)

    @pytest.mark.parametrize("minimum_length", [1, 2, 3, 4])
    def test_hysteresis(self, simple_linear_model, minimum_length):
        """ Test the minimum_event_length keyword of EventRecorder """
        m = simple_linear_model

        flow = np.zeros(len(m.timestepper))

        flow[:10] = [0, 0, 10, 0, 10, 10, 10, 0, 0, 0]
        # With min event length of 1. There are two events with lengths (1, 3)
        #                 |---|   |---------|
        # With min event length up to 4. There is one event with length 3
        #                         |---------|
        # Min event length >= 4 gives no events

        inpt = m.nodes['Input']
        inpt.max_flow = ArrayIndexedParameter(m, flow)

        # Force through whatever flow Input can provide
        otpt = m.nodes['Output']
        otpt.max_flow = 100
        otpt.cost = -100

        # Create the trigger using a threhsold parameter
        trigger = NodeThresholdRecorder(m, otpt, 4.0, predicate='>')
        evt_rec = EventRecorder(m, trigger, minimum_event_length=minimum_length)

        m.run()

        if minimum_length == 1:
            assert len(evt_rec.events) == 2
            assert_equal([1, 3], [e.duration for e in evt_rec.events])
        elif minimum_length < 4:
            assert len(evt_rec.events) == 1
            assert_equal([3, ], [e.duration for e in evt_rec.events])
        else:
            assert len(evt_rec.events) == 0

    @pytest.mark.parametrize("recorder_agg_func", ["min", "max", "mean", "median", "sum"])
    def test_statistic_recorder(self, cyclical_storage_model, recorder_agg_func):
        """ Test EventStatisticRecorder """
        m = cyclical_storage_model

        strg = m.nodes['Storage']
        inpt = m.nodes['Input']
        arry = NumpyArrayNodeRecorder(m, inpt)

        # Create the trigger using a threhsold parameter
        trigger = StorageThresholdRecorder(m, strg, 4.0, predicate='<=')
        evt_rec = EventRecorder(m, trigger, tracked_parameter=inpt.max_flow)
        evt_stat = EventStatisticRecorder(m, evt_rec, agg_func='max', event_agg_func='min', recorder_agg_func=recorder_agg_func)

        m.run()

        # Ensure there is at least one event
        assert evt_rec.events

        evt_values = {si.global_id:[] for si in m.scenarios.combinations}
        for evt in evt_rec.events:
            evt_values[evt.scenario_index.global_id].append(np.min(arry.data[evt.start.index:evt.end.index, evt.scenario_index.global_id]))

        func = TestEventRecorder.funcs[recorder_agg_func]

        agg_evt_values = []
        for k, v in sorted(evt_values.items()):
            if len(v) > 0:
                agg_evt_values.append(func(v))
            else:
                agg_evt_values.append(np.nan)

        # Test that the
        assert_allclose(evt_stat.values(), agg_evt_values)
        assert_allclose(evt_stat.aggregated_value(), np.max(agg_evt_values))


def test_progress_recorder(simple_linear_model):
    model = simple_linear_model
    rec = ProgressRecorder(model)
    model.run()


class TestHydroPowerRecorder:

    @pytest.mark.parametrize('efficiency', [1.0, 0.85])
    def test_constant_level(self, simple_storage_model, efficiency):
        """ Test HydropowerRecorder """
        m = simple_storage_model

        strg = m.nodes['Storage']
        otpt = m.nodes['Output']

        elevation = ConstantParameter(m, 100)
        rec = HydropowerRecorder(m, otpt, elevation, efficiency=efficiency)
        rec_total = TotalHydroEnergyRecorder(m, otpt, elevation, efficiency=efficiency)

        m.setup()
        m.step()

        # First step
        # Head: 100m
        # Flow: 8 m3/day
        # Power: 1000 * 9.81 * 8 * 100
        # Energy: power * 1 day = power
        np.testing.assert_allclose(rec.data[0, 0], 1000 * 9.81 * 8 * 100 * 1e-6 * efficiency)
        # Second step has the same answer in this model
        m.step()
        np.testing.assert_allclose(rec.data[1, 0], 1000 * 9.81 * 8 * 100 * 1e-6 * efficiency)
        np.testing.assert_allclose(rec_total.values()[0], 2 * 1000 * 9.81 * 8 * 100 * 1e-6 * efficiency)

    def test_varying_level(self, simple_storage_model):
        """ Test HydropowerRecorder with varying level on Storage node """
        from pywr.parameters import InterpolatedVolumeParameter
        m = simple_storage_model

        strg = m.nodes['Storage']
        otpt = m.nodes['Output']

        elevation = InterpolatedVolumeParameter(m, strg, [0, 10, 20], [0, 100, 200])
        rec = HydropowerRecorder(m, otpt, elevation)
        rec_total = TotalHydroEnergyRecorder(m, otpt, elevation)

        m.setup()
        m.step()

        # First step
        # Head: 100m
        # Flow: 8 m3/day
        # Power: 1000 * 9.81 * 8 * 100
        # Energy: power * 1 day = power
        np.testing.assert_allclose(rec.data[0, 0], 1000 * 9.81 * 8 * 100 * 1e-6)
        # Second step is at a lower head
        # Head: 70m
        m.step()
        np.testing.assert_allclose(rec.data[1, 0], 1000 * 9.81 * 8 * 70 * 1e-6)
        np.testing.assert_allclose(rec_total.values()[0], 1000 * 9.81 * 8 * 170 * 1e-6)

    def test_varying_level_with_turbine_level(self, simple_storage_model):
        """ Test HydropowerRecorder with varying level on Storage and defined level on the recorder """
        from pywr.parameters import InterpolatedVolumeParameter
        m = simple_storage_model

        strg = m.nodes['Storage']
        otpt = m.nodes['Output']

        elevation = InterpolatedVolumeParameter(m, strg, [0, 10, 20], [0, 100, 200])
        rec = HydropowerRecorder(m, otpt, elevation, turbine_elevation=80)
        rec_total = TotalHydroEnergyRecorder(m, otpt, elevation, turbine_elevation=80)

        m.setup()
        m.step()

        # First step
        # Head: 100 - 80 = 20m
        # Flow: 8 m3/day
        # Power: 1000 * 9.81 * 8 * 100
        # Energy: power * 1 day = power
        np.testing.assert_allclose(rec.data[0, 0], 1000 * 9.81 * 8 * 20 * 1e-6)
        # Second step is at a lower head
        # Head: 70 - 80: -10m (i.e. not sufficient)
        m.step()
        np.testing.assert_allclose(rec.data[1, 0], 1000 * 9.81 * 8 * 0 * 1e-6)
        np.testing.assert_allclose(rec_total.values()[0], 1000 * 9.81 * 8 * 20 * 1e-6)

    def test_load_from_json(self, ):
        """ Test example hydropower model loads and runs. """
        model = load_model("hydropower_example.json")

        r = model.recorders['turbine1_energy']

        # Check the recorder has loaded correctly
        assert r.water_elevation_parameter == model.parameters['reservoir1_level']
        assert r.node == model.nodes['turbine1']

        assert_allclose(r.turbine_elevation, 35.0)
        assert_allclose(r.efficiency, 0.85)
        assert_allclose(r.flow_unit_conversion, 1e3)

        # Finally, check model runs with the loaded recorder.
        model.run()


class TestGaussianKDEStorageRecorder:
    def test_kde_recorder(self, simple_storage_model):
        """A basic functional test of `GaussianKDEStorageRecorder`"""
        model = simple_storage_model
        res = model.nodes['Storage']

        kde = GaussianKDEStorageRecorder(model, res, target_volume_pc=0.2)

        model.run()

        pdf = kde.to_dataframe()
        p = kde.aggregated_value()
        assert pdf.shape == (101, 1)
        assert 0 < p < 1
        np.testing.assert_allclose(pdf.values, kde.values())

    def test_kde_from_json(self, simple_storage_model):
        """Test loading KDE recorder from JSON data."""
        model = simple_storage_model

        kde = load_recorder(model, {
            'type': 'GaussianKDEStorageRecorder',
            'node': 'Storage',
            'target_volume_pc': 0.2
        })

        model.run()

        pdf = kde.to_dataframe()
        p = kde.aggregated_value()
        assert pdf.shape == (101, 1)
        assert 0 < p < 1
        np.testing.assert_allclose(pdf.values, kde.values())

    def test_norm_kde_recorder(self, simple_storage_model):
        """A basic functional test of `NormalisedGaussianKDEStorageRecorder`"""
        model = simple_storage_model
        res = model.nodes['Storage']
        cc = ConstantParameter(model, 0.2)

        kde = NormalisedGaussianKDEStorageRecorder(model, res, parameter=cc)

        model.run()

        pdf = kde.to_dataframe()
        p = kde.aggregated_value()
        assert pdf.shape == (101, 1)
        assert 0 < p < 1
        np.testing.assert_allclose(pdf.values, kde.values())

    def test_norm_kde_from_json(self, simple_storage_model):
        """Test loading normalised KDE recorder from JSON data."""
        model = simple_storage_model
        cc = ConstantParameter(model, 0.2, name='my-parameter')

        kde = load_recorder(model, {
            'type': 'NormalisedGaussianKDEStorageRecorder',
            'node': 'Storage',
            'parameter': 'my-parameter'
        })

        model.run()

        pdf = kde.to_dataframe()
        p = kde.aggregated_value()
        assert pdf.shape == (101, 1)
        assert 0 < p < 1
        np.testing.assert_allclose(pdf.values, kde.values())



import os
import sys
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

def setup_package():
    compiler_directives = {
        "language_level": 3,
        "embedsignature": True,
    }

    define_macros = []

    compile_time_env = {
        "SOLVER_DEBUG": False,
    }

    annotate = False

    class new_build_ext(_build_ext):
        def finalize_options(self):
            # Defer the import of Cython and NumPy until after setup_requires
            from Cython.Build.Dependencies import cythonize
            import numpy

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                compiler_directives=compiler_directives,
                compile_time_env=compile_time_env,
                annotate=annotate,
            )
            if not self.include_dirs:
                self.include_dirs = []
            elif isinstance(self.include_dirs, str):
                self.include_dirs = [self.include_dirs]
            self.include_dirs.append(numpy.get_include())
            super().finalize_options()

    # Extra optional dependencies
    docs_extras = ["sphinx", "sphinx_rtd_theme", "numpydoc"]
    test_extras = ["pytest"]
    dev_extras = docs_extras + test_extras
    opt_extras = ["platypus-opt", "pygmo"]

    metadata = dict(
        name="pywr",
        description="Python Water Resource model",
        long_description=long_description(),
        long_description_content_type="text/x-rst",
        author="Joshua Arnott",
        author_email="josh@snorfalorpagus.net",
        url="https://github.com/pywr/pywr",
        setup_requires=["setuptools>=18.0", "setuptools_scm", "cython", "numpy"],
        install_requires=[
            "pandas",
            "networkx",
            "scipy",
            "tables",
            "openpyxl",
            "packaging",
            "matplotlib",
            "jinja2",
            "ipython"
        ],
        extras_require={
            "docs": docs_extras,
            "test": test_extras,
            "dev": dev_extras,
            "optimisation": opt_extras
        },
        cmdclass={"build_ext": new_build_ext},
        packages=[
            "pywr",
            "pywr.solvers",
            "pywr.domains",
            "pywr.parameters",
            "pywr.recorders",
            "pywr.notebook",
            "pywr.optimisation",
            "pywr.utils",
        ],
        package_data=package_data(),
        use_scm_version=True,
    )

    metadata["ext_modules"] = ext_modules = [
        Extension("pywr._core", ["pywr/_core.pyx"]),
        Extension("pywr._model", ["pywr/_model.pyx"]),
        Extension("pywr._component", ["pywr/_component.pyx"]),
        Extension("pywr.parameters._parameters", ["pywr/parameters/_parameters.pyx"]),
        Extension("pywr.parameters._polynomial", ["pywr/parameters/_polynomial.pyx"]),
        Extension("pywr.parameters._thresholds", ["pywr/parameters/_thresholds.pyx"]),
        Extension("pywr.parameters._control_curves", ["pywr/parameters/_control_curves.pyx"]),
        Extension("pywr.parameters._hydropower", ["pywr/parameters/_hydropower.pyx"]),
        Extension("pywr.parameters._activation_functions", ["pywr/parameters/_activation_functions.pyx"]),
        Extension("pywr.recorders._recorders", ["pywr/recorders/_recorders.pyx"]),
        Extension("pywr.recorders._thresholds", ["pywr/recorders/_thresholds.pyx"]),
        Extension("pywr.recorders._hydropower", ["pywr/recorders/_hydropower.pyx"]),
    ]

    config = parse_optional_arguments()

    if config["glpk"]:
        ext_modules.append(
                Extension("pywr.solvers.cython_glpk", ["pywr/solvers/cython_glpk.pyx"], libraries=["glpk"],)
        )

    if config["lpsolve"]:
        if os.name == "nt":
            lpsolve_macros = define_macros + [("WIN32", 1)]
        else:
            lpsolve_macros = define_macros
        ext_modules.append(
            Extension(
                "pywr.solvers.cython_lpsolve",
                ["pywr/solvers/cython_lpsolve.pyx"],
                libraries=["lpsolve55"],
                define_macros=lpsolve_macros,
            )
        )

    annotate = config["annotate"]

    if config["profile"]:
        compiler_directives["profile"] = True

    if config["trace"]:
        compiler_directives["linetrace"] = True
        define_macros.extend(
            [("CYTHON_TRACE", "1"), ("CYTHON_TRACE_NOGIL", "1"),]
        )

    if config["debug"]:
        compile_time_env["SOLVER_DEBUG"] = True

    setup(**metadata)


def long_description():
    with open("README.rst") as f:
        return f.read()


def package_data():
    pkg_data = {
        "pywr.notebook": ["*.js", "*.css", "*.html"],
        'pywr': ['*.pxd'],
        'pywr.parameters': ['*.pxd'],
        'pywr.recorders': ['*.pxd'],
        'pywr.solvers': ['*.pxd'],
    }
    if os.environ.get("PACKAGE_DATA", "false").lower() == "true":
        pkg_data["pywr"].extend([".libs/*", ".libs/licenses/*"])
    return pkg_data


def parse_optional_arguments():
    config = {
        "glpk": True,
        "lpsolve": True,
        "annotate": False,
        "profile": False,
        "trace": False,
        "debug": False,
    }

    if "--with-glpk" in sys.argv:
        config["glpk"] = True
        sys.argv.remove("--with-glpk")
    elif "PYWR_BUILD_GLPK" in os.environ:
        config["glpk"] = os.environ["PYWR_BUILD_GLPK"].lower() in ("true", "1",)
    elif "--without-glpk" in sys.argv:
        config["glpk"] = False
        sys.argv.remove("--without-glpk")

    if "--with-lpsolve" in sys.argv:
        config["lpsolve"] = True
        sys.argv.remove("--with-lpsolve")
    elif "PYWR_BUILD_LPSOLVE" in os.environ:
        config["lpsolve"] = os.environ["PYWR_BUILD_LPSOLVE"].lower() in ("true", "1",)
    elif "--without-lpsolve" in sys.argv:
        config["lpsolve"] = False
        sys.argv.remove("--without-lpsolve")

    if "--annotate" in sys.argv:
        config["annotate"] = True
        sys.argv.remove("--annotate")

    if "--enable-profiling" in sys.argv:
        config["profile"] = True
        sys.argv.remove("--enable-profiling")

    if "--enable-trace" in sys.argv:
        config["trace"] = True
        sys.argv.remove("--enable-trace")
    elif "PYWR_BUILD_TRACE" in os.environ:
        config["trace"] = os.environ["PYWR_BUILD_TRACE"].lower() in ("true", "1",)

    if "--enable-debug" in sys.argv:
        config["debug"] = True
        sys.argv.remove("--enable-debug")

    return config


if __name__ == "__main__":
    setup_package()



from libc.float cimport DBL_MAX
from libc.stdlib cimport malloc, free
from cython.view cimport array as cvarray
from pywr._core import BaseInput, BaseOutput, BaseLink
from pywr._core cimport *
from pywr.core import ModelStructureError
import time


cdef extern from "lpsolve/lp_lib.h":
    cdef struct _lprec:
        pass
    ctypedef _lprec lprec
    ctypedef unsigned char MYBOOL
    cdef unsigned char FALSE
    cdef unsigned char TRUE
    cdef unsigned char FR
    cdef unsigned char LE
    cdef unsigned char GE
    cdef unsigned char EQ
    cdef int OPTIMAL
    ctypedef double REAL
    cdef int CRITICAL
    cdef int FULL
    cdef int PRESOLVE_ROWS
    cdef int PRESOLVE_COLS
    cdef int PRESOLVE_LINDEP

    lprec* make_lp(int rows, int columns)
    MYBOOL resize_lp(lprec *lp, int rows, int columns)
    char* get_statustext(lprec *lp, int statuscode)
    MYBOOL get_variables(lprec *lp, REAL *var)
    MYBOOL get_ptr_variables(lprec *lp, REAL **var)

    MYBOOL add_constraint(lprec *lp, REAL *row, int constr_type, REAL rh)
    MYBOOL add_constraintex(lprec *lp, int count, REAL *row, int *colno, int constr_type, REAL rh)
    MYBOOL set_add_rowmode(lprec *lp, MYBOOL turnon)
    void delete_lp(lprec *lp)
    void free_lp(lprec **plp)
    void set_maxim(lprec *lp)
    void set_minim(lprec *lp)

    MYBOOL add_column(lprec *lp, REAL *column)
    MYBOOL add_columnex(lprec *lp, int count, REAL *column, int *rowno)

    MYBOOL set_obj(lprec *lp, int colnr, REAL value)
    MYBOOL set_bounds(lprec *lp, int colnr, REAL lower, REAL upper)
    MYBOOL set_constr_type(lprec *lp, int rownr, int con_type)
    MYBOOL set_bounds(lprec *lp, int colnr, REAL lower, REAL upper)
    MYBOOL set_lowbo(lprec *lp, int colnr, REAL value)
    MYBOOL set_row(lprec *lp, int rownr, REAL *row)
    MYBOOL set_rowex(lprec *lp, int rownr, int count, REAL *row, int *colno)
    MYBOOL set_rh(lprec *lp, int rownr, REAL value)
    MYBOOL set_rh_range(lprec *lp, int rownr, REAL deltavalue)
    int get_Norig_rows(lprec *lp)
    int get_Nrows(lprec *lp)
    int get_Lrows(lprec *lp)
    int get_nonzeros(lprec *lp)
    int get_Norig_columns(lprec *lp)
    int get_Ncolumns(lprec *lp)
    int solve(lprec *lp)
    void print_lp(lprec *lp)
    void set_verbose(lprec *lp, int verbose)
    void set_presolve(lprec *lp, int presolvemode, int maxloops)
    int get_presolve(lprec *lp)
    int get_presolveloops(lprec *lp)

cdef inline int set_row_bnds(lprec *prob, int row, double a, double b):
    if a == b:
        set_constr_type(prob, row, EQ)
        set_rh(prob, row, a)
        # set_rh_range(prob, row, 0.0)
    elif b == DBL_MAX:
        if a == -DBL_MAX:
            set_constr_type(prob, row, FR)
            set_rh(prob, row, 0.0)
            # set_rh_range(prob, row, 0.0)
        else:
            set_constr_type(prob, row, GE)
            set_rh(prob, row, a)
            # set_rh_range(prob, row, 0.0)
    elif a == -DBL_MAX:
        set_constr_type(prob, row, LE)
        set_rh(prob, row, b)
        # set_rh_range(prob, row, 0.0)
    else:
        set_constr_type(prob, row, LE)
        set_rh(prob, row, b)
        set_rh_range(prob, row, b-a)

cdef double inf = float('inf')

cdef inline double dbl_max_to_inf(double a):
    if a == DBL_MAX:
        return inf
    elif a == -DBL_MAX:
        return -inf
    return a

cdef inline double inf_to_dbl_max(double a):
    if a == inf:
        return DBL_MAX
    elif a == -inf:
        return -DBL_MAX
    return a

cdef class CythonLPSolveSolver:
    cdef lprec *prob
    cdef int idx_col_routes
    cdef int idx_col_demands
    cdef int idx_row_supplys
    cdef int idx_row_demands
    cdef int idx_row_cross_domain
    cdef int idx_row_storages
    cdef int idx_row_virtual_storages
    cdef int idx_row_aggregated
    cdef int idx_row_aggregated_min_max

    cdef public list routes
    cdef list supplys
    cdef list demands
    cdef list storages
    cdef list virtual_storages
    cdef list aggregated
    cdef public object stats
    cdef int num_routes
    cdef int num_scenarios
    cdef cvarray node_flows_arr
    cdef public cvarray route_flows_arr
    cdef public bint save_routes_flows

    def __cinit__(self):
        # create a new problem
        self.prob = make_lp(0, 0)
        if self.prob is NULL:
            raise MemoryError()
        set_verbose(self.prob, CRITICAL)

    def __dealloc__(self):
        if self.prob is not NULL:
            # free the problem
            delete_lp(self.prob)

    def __init__(self, save_routes_flows=False):
        self.stats = None
        self.save_routes_flows = save_routes_flows

    cpdef object setup(self, model):
        cdef Node supply
        cdef Node demand
        cdef Node node
        cdef AggregatedNode agg_node
        cdef double min_flow
        cdef double max_flow
        cdef double cost
        cdef double avail_volume
        cdef int col
        cdef int* ind
        cdef double* val
        cdef double lb
        cdef double ub
        cdef Timestep timestep
        cdef int status
        cdef cross_domain_col
        cdef MYBOOL ret
        cdef REAL *ptr_var

        if not model.graph.nodes():
            raise ModelStructureError("Model is empty")

        routes = model.find_all_routes(BaseInput, BaseOutput, valid=(BaseLink, BaseInput, BaseOutput))
        # Find cross-domain routes
        cross_domain_routes = model.find_all_routes(BaseOutput, BaseInput, max_length=2, domain_match='different')

        supplys = []
        demands = []
        storages = []
        virtual_storages = []
        aggregated = []
        aggregated_min_max = []
        for some_node in model.graph.nodes():
            if isinstance(some_node, (BaseInput, BaseLink)):
                supplys.append(some_node)
            if isinstance(some_node, BaseOutput):
                demands.append(some_node)
            if isinstance(some_node, VirtualStorage):
                virtual_storages.append(some_node)
            elif isinstance(some_node, Storage):
                storages.append(some_node)
            elif isinstance(some_node, AggregatedNode):
                if some_node.factors is not None:
                    # See discussion: https://github.com/pywr/pywr/pull/919
                    # Implementing fully dynamic factors in lpsolve is complicated.
                    if not some_node.has_fixed_factors:
                        raise ValueError("{} has one or more factors defined by a parameter. This is not allowed \
                                        when using the lpsolve solver. Please use the glpk or glpk-edge solver \
                                        instead".format(some_node.name))
                    aggregated.append(some_node)
                if some_node.min_flow > -inf or \
                   some_node.max_flow < inf:
                    aggregated_min_max.append(some_node)

        if len(routes) == 0:
            raise ModelStructureError("Model has no valid routes")
        if (len(supplys) + len(demands)) == 0:
            raise ModelStructureError("Model has no non-storage nodes")

        # clear the previous problem
        ret = resize_lp(self.prob, 0, 0)
        if ret == FALSE:
            return -1
        set_minim(self.prob)

        self.num_routes = len(routes)
        self.num_scenarios = len(model.scenarios.combinations)

        if self.save_routes_flows:
            # If saving flows this array needs to be 2D (one for each scenario)
            self.route_flows_arr = cvarray(shape=(self.num_scenarios, self.num_routes),
                                           itemsize=sizeof(double), format="d")
        else:
            # Otherwise the array can just be used to store a single solve to save some memory
            self.route_flows_arr = cvarray(shape=(self.num_routes, ), itemsize=sizeof(double), format="d")

        # add a column for each route and demand
        self.idx_col_routes = get_Norig_columns(self.prob)+1
        self.idx_col_demands = self.idx_col_routes + len(routes)
        ret = resize_lp(self.prob, 0, len(routes)+len(demands))
        ret = set_add_rowmode(self.prob, TRUE)

        # create a lookup for the cross-domain routes.
        cross_domain_cols = {}
        for cross_domain_route in cross_domain_routes:
            # These routes are only 2 nodes. From demand to supply
            demand, supply = cross_domain_route
            # TODO make this time varying.
            conv_factor = supply.get_conversion_factor()
            supply_cols = [(n, conv_factor) for n, route in enumerate(routes) if route[0] is supply]
            # create easy lookup for the route columns this demand might
            # provide cross-domain connection to
            if demand in cross_domain_cols:
                cross_domain_cols[demand].extend(supply_cols)
            else:
                cross_domain_cols[demand] = supply_cols

        # explicitly set bounds on route and demand columns
        for col, route in enumerate(routes):
            ret = add_columnex(self.prob, 0, NULL, NULL)
            ret = set_lowbo(self.prob, self.idx_col_routes+col, 0.0)
        for col, demand in enumerate(demands):
            ret = add_columnex(self.prob, 0, NULL, NULL)
            ret = set_lowbo(self.prob, self.idx_col_demands+col, 0.0)

        # constrain supply minimum and maximum flow
        self.idx_row_supplys = get_Norig_rows(self.prob)+1
        ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(supplys), get_Norig_columns(self.prob))
        for col, supply in enumerate(supplys):
            # TODO is this a bit hackish??
            if isinstance(supply, BaseInput):
                cols = [n for n, route in enumerate(routes) if route[0] is supply]
            else:
                cols = [n for n, route in enumerate(routes) if supply in route]

            ind = <int*>malloc(len(cols) * sizeof(int))
            val = <double*>malloc(len(cols) * sizeof(double))
            for n, c in enumerate(cols):
                ind[n] = 1+c
                val[n] = 1
            ret = add_constraintex(self.prob, len(cols), val, ind, GE, 0.0)

            # set_rowex(self.prob, self.idx_row_supplys+col, len(cols)+1, val, ind)
            # set_row_bnds(self.prob, self.idx_row_supplys+col, 0.0, 0.0)

            free(ind)
            free(val)

        # link supply and demand variables
        self.idx_row_demands = get_Norig_rows(self.prob)+1
        ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(demands), get_Norig_columns(self.prob))
        if len(cross_domain_cols) > 0:
            self.idx_row_cross_domain = get_Norig_rows(self.prob)+1
            ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(cross_domain_cols), get_Norig_columns(self.prob))
        cross_domain_col = 0
        for col, demand in enumerate(demands):
            cols = [n for n, route in enumerate(routes) if route[-1] is demand]
            ind = <int*>malloc((1+len(cols)) * sizeof(int))
            val = <double*>malloc((1+len(cols)) * sizeof(double))
            for n, c in enumerate(cols):
                ind[n] = 1+c
                val[n] = 1
            ind[len(cols)] = self.idx_col_demands+col
            val[len(cols)] = -1
            ret = add_constraintex(self.prob, len(cols)+1, val, ind, EQ, 0.0)
            free(ind)
            free(val)

        for col, demand in enumerate(demands):
            # Add constraint for cross-domain routes
            # i.e. those from a demand to a supply
            if demand in cross_domain_cols:
                col_vals = cross_domain_cols[demand]
                ind = <int*>malloc((1+len(col_vals)) * sizeof(int))
                val = <double*>malloc((1+len(col_vals)) * sizeof(double))
                for n, (c, v) in enumerate(col_vals):
                    ind[n] = 1+c
                    val[n] = 1./v
                ind[len(col_vals)] = self.idx_col_demands+col
                val[len(col_vals)] = -1
                add_constraintex(self.prob, len(col_vals)+1, val, ind, EQ, 0.0)
                cross_domain_col += 1
                free(ind)
                free(val)

        # storage
        if len(storages):
            self.idx_row_storages = get_Norig_rows(self.prob)+1
            ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(storages), get_Norig_columns(self.prob))
        for col, storage in enumerate(storages):
            cols_output = [n for n, demand in enumerate(demands) if demand in storage.outputs]
            cols_input = [n for n, route in enumerate(routes) if route[0] in storage.inputs]
            ind = <int*>malloc((len(cols_output)+len(cols_input)) * sizeof(int))
            val = <double*>malloc((len(cols_output)+len(cols_input)) * sizeof(double))
            for n, c in enumerate(cols_output):
                ind[n] = self.idx_col_demands+c
                val[n] = 1
            for n, c in enumerate(cols_input):
                ind[len(cols_output)+n] = self.idx_col_routes+c
                val[len(cols_output)+n] = -1
            add_constraintex(self.prob, len(cols_output)+len(cols_input), val, ind, EQ, 0.0)
            free(ind)
            free(val)

        if len(virtual_storages):
            self.idx_row_virtual_storages = get_Norig_rows(self.prob) + 1
            ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(virtual_storages), get_Norig_columns(self.prob))
        for col, storage in enumerate(virtual_storages):
            # We need to handle the same route appearing twice here.
            cols = {}
            for n, route in enumerate(routes):
                for some_node in route:
                    try:
                        i = storage.nodes.index(some_node)
                    except ValueError:
                        pass
                    else:
                        try:
                            cols[n] += storage.factors[i]
                        except KeyError:
                            cols[n] = storage.factors[i]

            ind = <int*>malloc((len(cols)) * sizeof(int))
            val = <double*>malloc((len(cols)) * sizeof(double))
            for n, (c, f) in enumerate(cols.items()):
                ind[n] = self.idx_col_routes+c
                val[n] = -f

            add_constraintex(self.prob, len(cols), val, ind, EQ, 0.0)

            free(ind)
            free(val)

        # aggregated node flow ratio constraints
        if len(aggregated):
            self.idx_row_aggregated = get_Norig_rows(self.prob)+1

        for agg_node in aggregated:
            nodes = agg_node.nodes
            # NB this only works because the solver supports only fixed (i.e. ConstantParameter) factors
            factors = [f.get_double_variables()[0] for f in agg_node.factors]
            assert(len(nodes) == len(factors))

            ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(agg_node.nodes)-1, get_Norig_columns(self.prob))
            row = get_Norig_rows(self.prob)+1

            cols = []
            for node in nodes:
                cols.append([n for n, route in enumerate(routes) if node in route])

            # normalise factors
            f0 = factors[0]
            factors_norm = [f0/f for f in factors]

            # update matrix
            for n in range(len(nodes)-1):
                length = len(cols[0])+len(cols[n+1])
                ind = <int*>malloc(length * sizeof(int))
                val = <double*>malloc(length * sizeof(double))
                for i, c in enumerate(cols[0]):
                    ind[i] = 1+c
                    val[i] = 1.0
                for i, c in enumerate(cols[n+1]):
                    ind[len(cols[0])+i] = 1+c
                    val[len(cols[0])+i] = -factors_norm[n+1]

                add_constraintex(self.prob, length, val, ind, EQ, 0.0)

                free(ind)
                free(val)

        # aggregated node min/max flow constraints
        if aggregated_min_max:
            self.idx_row_aggregated_min_max = get_Norig_rows(self.prob)+1
            ret = resize_lp(self.prob, get_Norig_rows(self.prob)+len(aggregated_min_max), get_Norig_columns(self.prob))

        for row, agg_node in enumerate(aggregated_min_max):
            row = self.idx_row_aggregated_min_max + row
            nodes = agg_node.nodes
            min_flow = agg_node.min_flow
            max_flow = agg_node.max_flow
            if min_flow is None:
                min_flow = -inf
            if max_flow is None:
                max_flow = inf
            min_flow = inf_to_dbl_max(min_flow)
            max_flow = inf_to_dbl_max(max_flow)

            weights = agg_node.flow_weights
            if weights is None:
                weights = [1.0]*len(nodes)

            matrix = {}
            for node, w in zip(nodes, weights):
                for n, route in enumerate(routes):
                    if node in route:
                        matrix[n] = w

            length = len(matrix)
            ind = <int*>malloc(length * sizeof(int))
            val = <double*>malloc(length * sizeof(double))
            for i, col in enumerate(matrix):
                ind[i] = 1+col
                val[i] = matrix[col]
            add_constraintex(self.prob, length, val, ind, EQ, 0.0)
            set_row_bnds(self.prob, row, min_flow, max_flow)
            free(ind)
            free(val)

        # reset stats
        self.stats = {
            'total': 0.0,
            'lp_solve': 0.0,
            'result_update': 0.0,
            'bounds_update_routes': 0.0,
            'bounds_update_nonstorage': 0.0,
            'bounds_update_storage': 0.0,
            'number_of_rows': get_Nrows(self.prob),
            'number_of_cols': get_Ncolumns(self.prob),
            'number_of_nonzero': get_nonzeros(self.prob),
            'number_of_routes': len(routes),
            'number_of_nodes': len(model.nodes)
        }

        ret = set_add_rowmode(self.prob, FALSE)
        self.routes = routes
        self.supplys = supplys
        self.demands = demands
        self.storages = storages
        self.virtual_storages = virtual_storages
        self.aggregated = aggregated

    cpdef object solve(self, model):
        cdef int[:] scenario_combination
        cdef int scenario_id
        cdef ScenarioIndex scenario_index

        for scenario_index in model.scenarios.combinations:
            self._solve_scenario(model, scenario_index)

    cdef object _solve_scenario(self, model, ScenarioIndex scenario_index):
        cdef Node supply
        cdef Node demand
        cdef Node node
        cdef double min_flow
        cdef double max_flow
        cdef double cost
        cdef double avail_volume
        cdef double t0
        cdef int col
        cdef int* ind
        cdef double* val
        cdef double lb
        cdef double ub
        cdef Timestep timestep
        cdef int status
        cdef cross_domain_col
        cdef MYBOOL ret
        cdef REAL *ptr_var

        timestep = model.timestep

        routes = self.routes
        supplys = self.supplys
        demands = self.demands
        storages = self.storages
        virtual_storages = self.virtual_storages

        t0 = time.perf_counter()

        # update route properties
        for col, route in enumerate(routes):
            cost = route[0].get_cost(scenario_index)
            for node in route[1:-1]:
                if isinstance(node, BaseLink):
                    cost += node.get_cost(scenario_index)
            set_obj(self.prob, self.idx_col_routes+col, cost)

        self.stats['bounds_update_routes'] += time.perf_counter() - t0
        t0 = time.perf_counter()

        # update supply properties
        for col, supply in enumerate(supplys):
            min_flow = inf_to_dbl_max(supply.get_min_flow(scenario_index))
            max_flow = inf_to_dbl_max(supply.get_max_flow(scenario_index))
            set_row_bnds(self.prob, self.idx_row_supplys+col, min_flow, max_flow)

        # update demand properties
        for col, demand in enumerate(demands):
            min_flow = inf_to_dbl_max(demand.get_min_flow(scenario_index))
            max_flow = inf_to_dbl_max(demand.get_max_flow(scenario_index))
            cost = demand.get_cost(scenario_index)
            set_bounds(self.prob, self.idx_col_demands+col, min_flow, max_flow)
            set_obj(self.prob, self.idx_col_demands+col, cost)

        self.stats['bounds_update_nonstorage'] += time.perf_counter() - t0
        t0 = time.perf_counter()

        # update storage node constraint
        for col, storage in enumerate(storages):
            max_volume = storage.get_max_volume(scenario_index)
            avail_volume = max(storage._volume[scenario_index.global_id] - storage.get_min_volume(scenario_index), 0.0)
            # change in storage cannot be more than the current volume or
            # result in maximum volume being exceeded
            lb = -avail_volume/timestep.days
            ub = (max_volume - storage._volume[scenario_index.global_id]) / timestep.days
            set_row_bnds(self.prob, self.idx_row_storages+col, lb, ub)

        # update virtual storage node constraint
        for col, storage in enumerate(virtual_storages):
            if not storage.active:
                set_row_bnds(self.prob, self.idx_row_virtual_storages+col, -DBL_MAX, DBL_MAX)
            else:
                max_volume = storage.get_max_volume(scenario_index)
                avail_volume = max(storage._volume[scenario_index.global_id] - storage.get_min_volume(scenario_index), 0.0)
                # change in storage cannot be more than the current volume or
                # result in maximum volume being exceeded
                lb = -avail_volume/timestep.days
                ub = (max_volume - storage._volume[scenario_index.global_id]) / timestep.days
                set_row_bnds(self.prob, self.idx_row_virtual_storages+col, lb, ub)

        self.stats['bounds_update_storage'] += time.perf_counter() - t0
        t0 = time.perf_counter()
        # attempt to solve the linear programme

        # print_lp(self.prob)
        # set_presolve(self.prob, PRESOLVE_ROWS | PRESOLVE_COLS, get_presolveloops(self.prob))
        # attempt to solve the linear programme
        status = solve(self.prob)

        if status != OPTIMAL:
            raise RuntimeError(get_statustext(self.prob, status))

        self.stats['lp_solve'] += time.perf_counter() - t0
        t0 = time.perf_counter()

        get_ptr_variables(self.prob, &ptr_var)

        cdef double[:] route_flows
        if self.save_routes_flows:
            route_flows = self.route_flows_arr[scenario_index.global_id, :]
        else:
            route_flows = self.route_flows_arr

        for col in range(0, len(routes)):
            route_flows[col] = ptr_var[col]

        change_in_storage = []

        result = {}

        for col, route in enumerate(routes):
            flow = route_flows[col]
            # TODO make this cleaner.
            route[0].commit(scenario_index.global_id, flow)
            route[-1].commit(scenario_index.global_id, flow)
            for node in route[1:-1]:
                if isinstance(node, BaseLink):
                    node.commit(scenario_index.global_id, flow)

        self.stats['result_update'] += time.perf_counter() - t0

        return route_flows, change_in_storage



"""
This module contains a set of pywr._core.Parameter subclasses for defining control curve based parameters.
"""
from ._control_curves import (ControlCurveParameter, BaseControlCurveParameter,
    ControlCurveInterpolatedParameter, ControlCurveIndexParameter, ControlCurvePiecewiseInterpolatedParameter)



import os
from numpy.testing import assert_allclose
import pywr.core
from pywr.core import Model, Timestep
import pandas


def load_model(filename=None, data=None, solver=None, model_klass=None):
    '''Load a test model and check it'''
    if data is None:
        path = os.path.join(os.path.dirname(__file__), 'models')
        with open(os.path.join(path, filename), 'r') as f:
            data = f.read()
    else:
        path = None

    if model_klass is None:
        model_klass = Model

    model = model_klass.loads(data, path=path, solver=solver)
    model.check()
    return model


def assert_model(model, expected_node_results):
    __tracebackhide__ = True
    model.step()

    for node in model.nodes:
        if node.name in expected_node_results:
            if isinstance(node, pywr.core.BaseNode):
                assert_allclose(expected_node_results[node.name], node.flow, atol=1e-7)
            elif isinstance(node, pywr.core.Storage):
                assert_allclose(expected_node_results[node.name], node.volume, atol=1e-7)

def build_timestep(model, date):
    dt = pandas.to_datetime(date)
    timestep = Timestep(dt, (dt - model.timestepper.start).days, 1)
    return timestep



import os
import pytest
from pywr import hashes


TEST_FOLDER = os.path.dirname(__file__)

@pytest.mark.parametrize("filename, algorithm, hash, correct",[
    ('timeseries2.csv', 'md5', 'a5c4032e2d8f5205ca99dedcfa4cd18e', True),
    ('timeseries2.csv', 'sha256', '0f75b3cee325d37112687d3d10596f44e0add374f4e40a1b6687912c05e65366', True),
    ('timeseries2.h5', 'md5', '0f6c65a36851c89c7c4e63ab1893554b', True),
    # This next one is the sha256 hash, but is given as md5. Therefore it should fail.
    ('timeseries2.h5', 'md5', '1272702d60694f3417b910fb158e717de4fccdbf6aa10aa37f1c95cd78f8075e', False),
                         ])
def test_hash_timeseries2_(filename, algorithm, hash, correct):
    """ Test the hash value of files in the models directory """

    fullname = os.path.join(TEST_FOLDER, 'models', filename)

    if correct:
        hashes.check_hash(fullname, hash, algorithm=algorithm)
    else:
        with pytest.raises(hashes.HashMismatchError):
            hashes.check_hash(fullname, hash, algorithm=algorithm)



cimport numpy as np
import numpy as np


cdef enum Predicates:
    LT = 0
    GT = 1
    EQ = 2
    LE = 3
    GE = 4
_predicate_lookup = {
    "LT": Predicates.LT, "<": Predicates.LT,
    "GT": Predicates.GT, ">": Predicates.GT,
    "EQ": Predicates.EQ, "=": Predicates.EQ,
    "LE": Predicates.LE, "<=": Predicates.LE,
    "GE": Predicates.GE, ">=": Predicates.GE,
}

cdef int compare(double x, double threshold, int predicate) except? -1:
    """Returns 1 if the predicate evalutes True, else 0"""

    if predicate == Predicates.LT:
        ind = x < threshold
    elif predicate == Predicates.GT:
        ind = x > threshold
    elif predicate == Predicates.LE:
        ind = x <= threshold
    elif predicate == Predicates.GE:
        ind = x >= threshold
    else:
        ind = x == threshold
    return ind


cdef class StorageThresholdRecorder(StorageRecorder):
    """ Recorder for tracking state of Storage volume against a threshold.

    Parameters
    ----------
    node : Storage
        Storage instance to compare with the threshold.
    threshold : double
        Threshold to compare the value of the recorder to
    values : iterable of doubles
        If the predicate evaluates False the zeroth value is returned,
        otherwise the first value is returned.
    predicate : string
        One of {"LT", "GT", "EQ", "LE", "GE"}.


    """
    def __init__(self, model, Storage node, threshold, *args,  predicate=None, **kwargs):
        super(StorageThresholdRecorder, self).__init__(model, node, *args, **kwargs)
        self.threshold = threshold

        if predicate is None:
            predicate = Predicates.LT
        elif isinstance(predicate, str):
            predicate = _predicate_lookup[predicate.upper()]
        self.predicate = predicate

    cpdef setup(self):
        self._state = np.zeros(len(self.model.scenarios.combinations), dtype=np.int32)

    cpdef reset(self):
        self._state[...] = 0

    cpdef double[:] values(self) except *:
        return np.array(self._state, dtype=np.float64)

    cpdef after(self):
        cdef double volume
        cdef ScenarioIndex scenario_index
        for scenario_index in self.model.scenarios.combinations:
            volume = self._node._volume[scenario_index.global_id]
            self._state[scenario_index.global_id] = compare(volume, self.threshold, self.predicate)
        return 0


cdef class NodeThresholdRecorder(NodeRecorder):
    """ Recorder for tracking state of Node flow against a threshold.

    Parameters
    ----------
    node : Node
        Node instance to compare with the threshold.
    threshold : double
        Threshold to compare the value of the recorder to
    values : iterable of doubles
        If the predicate evaluates False the zeroth value is returned,
        otherwise the first value is returned.
    predicate : string
        One of {"LT", "GT", "EQ", "LE", "GE"}.


    """
    def __init__(self, model, AbstractNode node, threshold, *args,  predicate=None, **kwargs):
        super(NodeThresholdRecorder, self).__init__(model, node, *args, **kwargs)
        self.threshold = threshold

        if predicate is None:
            predicate = Predicates.LT
        elif isinstance(predicate, str):
            predicate = _predicate_lookup[predicate.upper()]
        self.predicate = predicate

    cpdef setup(self):
        self._state = np.zeros(len(self.model.scenarios.combinations), dtype=np.int32)

    cpdef reset(self):
        self._state[...] = 0

    cpdef double[:] values(self) except *:
        return np.array(self._state, dtype=np.float64)

    cpdef after(self):
        cdef double flow
        cdef ScenarioIndex scenario_index
        for scenario_index in self.model.scenarios.combinations:
            flow = self._node._flow[scenario_index.global_id]
            self._state[scenario_index.global_id] = compare(flow, self.threshold, self.predicate)
        return 0






from pywr.dataframe_tools import align_and_resample_dataframe, ResamplingError
import numpy as np
import pandas as pd
import pytest


def make_df(freq, start='2015-01-01', end='2015-12-31'):
    # Daily time-step
    index = pd.period_range(start, end, freq=freq)
    series = pd.DataFrame(np.arange(len(index), dtype=np.float64), index=index)
    return series


def make_model_index(freq, start='2015-01-01', end='2015-12-31'):
    return pd.period_range(start, end, freq=freq)


class TestDownSampling:
    """Test for down-sampling a dataframe to lower frequency model time-step."""

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_daily_to_monthly(self, resample_func):
        """Test daily data to monthly model time-step."""
        input_df = make_df('D')
        model_index = make_model_index('M')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, input_resampled.resample('M').agg(resample_func))

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_daily_to_weekly(self, resample_func):
        """Test daily to weekly model time-step."""
        input_df = make_df('D')
        model_index = make_model_index('W')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, input_resampled.resample('W').agg(resample_func))

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_daily_to_7daily(self, resample_func):
        """Test daily to 7-day model time-step."""
        input_df = make_df('D')
        model_index = make_model_index('7D')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, input_resampled.resample('7D').agg(resample_func))

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_misaligned_daily_to_7daily(self, resample_func):
        """Test daily to 7-day model time-step."""
        input_df = make_df('D', start='2014-12-20')
        model_index = make_model_index('7D')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, input_resampled.resample('7D').agg(resample_func))

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_weekly_to_monthly(self, resample_func):
        """Test weekly to monthly model time-step."""
        input_df = make_df('W')
        model_index = make_model_index('M')

        with pytest.raises(ResamplingError):
            input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_weekly_to_two_weekly(self, resample_func):
        """Test weekly to two-weekly model time-step."""
        input_df = make_df('W')
        model_index = make_model_index('2W')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, input_resampled.resample('2W').agg(resample_func))


class TestUpSampling:
    """Test for up-sampling a dataframe to higher frequency model time-step."""

    def test_monthly_to_daily(self):
        """Test monthly data to daily model time-step."""
        input_df = make_df('M')
        model_index = make_model_index('D')

        input_resampled = align_and_resample_dataframe(input_df, model_index)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    def test_weekly_to_daily(self):
        """Test weekly aligned to daily model time-step."""
        # Note that we have to choose a week starting on Wednesday because 2015-01-01 is a Wednesday.
        input_df = make_df('W-WED')
        model_index = make_model_index('D')

        input_resampled = align_and_resample_dataframe(input_df, model_index)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    def test_7daily_to_daily(self):
        """Test weekly aligned to daily model time-step."""
        # Note that we have to choose a week starting on Wednesday because 2015-01-01 is a Wednesday.
        input_df = make_df('7D')
        model_index = make_model_index('D')

        input_resampled = align_and_resample_dataframe(input_df, model_index)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    def test_nonaligned_weekly_to_daily(self):
        """Test weekly non-aligned data to daily model time-step."""
        input_df = make_df('W')
        model_index = make_model_index('D')

        input_resampled = align_and_resample_dataframe(input_df, model_index)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_weekly_to_7daily(self, resample_func):
        """Test weekly to 7-day model time-step."""
        # Note that we have to choose a week starting on Wednesday because 2015-01-01 is a Wednesday.
        input_df = make_df('W-WED')
        model_index = make_model_index('7D')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']
        expected_df = expected_df.resample('7D').agg(resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    @pytest.mark.parametrize('resample_func', ['mean', 'max'])
    def test_nonaligned_weekly_to_7daily(self, resample_func):
        """Test invalid weekly data to 7-day model time-step."""
        # Because the weeks do not align with the 7-day time-step it is not possible to align this data.
        input_df = make_df('W')
        model_index = make_model_index('7D')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func=resample_func)

        pd.testing.assert_index_equal(input_resampled.index, model_index)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']
        expected_df = expected_df.resample('7D').agg(resample_func)

        pd.testing.assert_frame_equal(input_resampled, expected_df)

    def test_annual_to_daily(self):

        input_df = make_df('A', '2010-01-01', '2020-01-01')
        model_index = make_model_index('D', '2010-01-01', '2020-01-01')

        input_resampled = align_and_resample_dataframe(input_df, model_index)

        expected_df = input_df.resample('D').ffill()
        expected_df = expected_df['2010-01-01':'2020-01-01']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    def test_annual_to_monthly(self):

        input_df = make_df('A', '2010-01-01', '2020-01-01')
        model_index = make_model_index('M', '2010-01-01', '2020-01-01')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func='ffill')

        expected_df = input_df.resample('M').ffill()
        expected_df = expected_df['2010-01-01':'2020-01-01']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)

    def test_two_weekly_to_weekly(self):
        """Test weekly to two-weekly model time-step."""
        input_df = make_df('2W')
        model_index = make_model_index('W')

        input_resampled = align_and_resample_dataframe(input_df, model_index, resample_func='ffill')

        expected_df = input_df.resample('W').ffill()
        expected_df = expected_df['2015-01-01':'2015-12-31']

        pd.testing.assert_index_equal(input_resampled.index, model_index)
        pd.testing.assert_frame_equal(input_resampled, expected_df)



# -*- coding: utf-8 -*-
"""
A series of tests of the Scenario objects and associated infrastructure


"""
from pywr.core import Model, Input, Output, Link, Storage, Scenario
from pywr.parameters import ConstantScenarioParameter
from pywr.recorders import NumpyArrayStorageRecorder, NumpyArrayNodeRecorder
from pywr.hashes import HashMismatchError
from helpers import assert_model, load_model
from fixtures import simple_linear_model
import numpy as np
from numpy.testing import assert_equal, assert_allclose
import pytest


def test_scenario_collection():
    """ Basic test of Scenario and ScenarioCollection API """

    model = Model()

    # There is 1 combination when there are no Scenarios
    model.scenarios.setup()
    assert(len(model.scenarios.combinations) == 1)
    assert(len(model.scenarios) == 0)
    scA = Scenario(model, 'Scenario A', size=3)
    model.scenarios.setup()
    assert(len(model.scenarios.combinations) == 3)
    assert(len(model.scenarios) == 1)
    scA = Scenario(model, 'Scenario B', size=2)
    model.scenarios.setup()
    assert(len(model.scenarios.combinations) == 6)
    assert(len(model.scenarios) == 2)

    assert_equal([comb.indices for comb in model.scenarios.combinations],
                 [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]])

    names = model.scenarios.combination_names
    for n, (ia, ib) in zip(names, [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]):
        assert n == 'Scenario A.{:03d}-Scenario B.{:03d}'.format(ia, ib)

    index = model.scenarios.multiindex
    assert_equal(index.tolist(),
                 [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]])
    assert_equal(index.names, ['Scenario A', 'Scenario B'])


def test_scenario(simple_linear_model, ):
    """Basic test of Scenario functionality"""
    model = simple_linear_model  # Convenience renaming

    scenario = Scenario(model, 'Inflow', size=2)
    model.nodes["Input"].max_flow = ConstantScenarioParameter(model, scenario, [5.0, 10.0])

    model.nodes["Output"].max_flow = 5.0
    model.nodes["Output"].cost = -2.0

    expected_node_results = {
        "Input": [5.0, 5.0],
        "Link": [5.0, 5.0],
        "Output": [5.0, 5.0],
    }

    assert_model(model, expected_node_results)


def test_two_scenarios(simple_linear_model, ):
    """Basic test of Scenario functionality"""
    model = simple_linear_model  # Convenience renaming

    scenario_input = Scenario(model, 'Inflow', size=2)
    model.nodes["Input"].max_flow = ConstantScenarioParameter(model, scenario_input, [5.0, 10.0])

    scenario_outflow = Scenario(model, 'Outflow', size=2, ensemble_names=['High', 'Low'])
    model.nodes["Output"].max_flow = ConstantScenarioParameter(model, scenario_outflow, [3.0, 8.0])
    model.nodes["Output"].cost = -2.0

    # Check ensemble names are provided in the multi-index
    index = model.scenarios.multiindex
    assert index.levels[0].name == 'Inflow'
    assert index.levels[1].name == 'Outflow'
    assert np.all(index.levels[1] == ['High', 'Low'])

    # add numpy recorders to input and output nodes
    NumpyArrayNodeRecorder(model, model.nodes["Input"], name="input")
    NumpyArrayNodeRecorder(model, model.nodes["Output"], name="output")

    expected_node_results = {
        "Input": [3.0, 5.0, 3.0, 8.0],
        "Link": [3.0, 5.0, 3.0, 8.0],
        "Output": [3.0, 5.0, 3.0, 8.0],
    }

    assert_model(model, expected_node_results)

    model.run()

    # combine recorder outputs to a single dataframe
    df = model.to_dataframe()
    assert(df.shape == (365, 2 * 2 * 2))
    assert_allclose(df["input", 0, 'High'].iloc[0], 3.0)
    assert_allclose(df["input", 0, 'Low'].iloc[0], 5.0)
    assert_allclose(df["input", 1, 'High'].iloc[0], 3.0)
    assert_allclose(df["input", 1, 'Low'].iloc[0], 8.0)


def test_scenario_two_parameter(simple_linear_model, ):
    """Basic test of Scenario functionality"""
    model = simple_linear_model  # Convenience renaming

    scenario_input = Scenario(model, 'Inflow', size=2)
    model.nodes["Input"].max_flow = ConstantScenarioParameter(model, scenario_input, [5.0, 10.0])

    model.nodes["Output"].max_flow = ConstantScenarioParameter(model, scenario_input, [8.0, 3.0])
    model.nodes["Output"].cost = -2.0

    expected_node_results = {
        "Input": [5.0, 3.0],
        "Link": [5.0, 3.0],
        "Output": [5.0, 3.0],
    }

    assert_model(model, expected_node_results)


def test_scenario_storage():
    """Test the behaviour of Storage nodes with multiple scenarios

    The model defined has two inflow scenarios: 5 and 10. It is expected that
    the volume in the storage node should increase at different rates in the
    two scenarios.
    """
    model = Model()

    i = Input(model, 'input', max_flow=999)
    s = Storage(model, 'storage', inputs=1, outputs=1, max_volume=1000, initial_volume=500)
    o = Output(model, 'output', max_flow=999)

    scenario_input = Scenario(model, 'Inflow', size=2)
    i.min_flow = ConstantScenarioParameter(model, scenario_input, [5.0, 10.0])

    i.connect(s)
    s.connect(o)

    s_rec = NumpyArrayStorageRecorder(model, s)

    model.run()

    assert_allclose(i.flow, [5, 10])
    assert_allclose(s_rec.data[0], [505, 510])
    assert_allclose(s_rec.data[1], [510, 520])


@pytest.mark.parametrize("json_file", ['simple_with_scenario.json', 'simple_with_scenario_wrapper.json'])
def test_scenarios_from_json(json_file):
    """
    Test a simple model with two scenarios.

    The model varies in the inflow by "scenario A" and the demand
    by "scenario B". The test ensures the correct size of model is
    created, and uses a `NumpyArrayNodeRecorder` to check the output
    in multiple dimensions is correct. The latter is done using
    the `MultiIndex` on the `DataFrame` from the recorder.
    """

    model = load_model(json_file)
    assert len(model.scenarios) == 2

    model.setup()
    assert len(model.scenarios.combinations) == 20
    model.run()

    # Test the recorder data is correct
    df = model.recorders['demand1'].to_dataframe()

    assert df.shape[1] == 20
    assert df.columns.names[0] == 'scenario A'
    assert_equal(df.columns.levels[0], np.arange(10))
    assert df.columns.names[1] == 'scenario B'
    assert_equal(df.columns.levels[1], np.array(['First', 'Second']))
    # Data for first demand (B) ensemble
    d1 = df.xs('First', level='scenario B', axis=1).iloc[0, :].values
    assert_allclose(d1, [10]*10)
    # Data for second demand (B) ensemble
    d2 = df.xs('Second', level='scenario B', axis=1).iloc[0, :]
    assert_allclose(d2, [10, 11, 12, 13, 14]+[15]*5)


def test_timeseries_with_scenarios():

    model = load_model('timeseries2.json')

    model.setup()

    assert len(model.scenarios) == 1

    model.step()
    catchment1 = model.nodes['catchment1']

    step1 = np.array([21.64, 21.72, 23.97, 23.35, 21.79, 21.52, 21.21, 22.58, 26.19, 25.71], dtype=np.float64)
    assert_allclose(catchment1.flow, step1)

    model.step()
    step2 = np.array([20.03, 20.10, 22.18, 21.62, 20.17, 19.92, 19.63, 20.90, 24.24, 23.80], dtype=np.float64)
    # Low tolerance because test values were truncated to 2 decimal places.
    assert_allclose(catchment1.flow, step2)

    model.finish()


def test_timeseries_with_scenarios_hdf():
    # this test uses TablesArrayParameter
    model = load_model('timeseries2_hdf.json')

    model.setup()

    assert len(model.scenarios) == 1

    catchment1 = model.nodes['catchment1']

    model.step()
    step1 = np.array([21.64, 21.72, 23.97, 23.35, 21.79, 21.52, 21.21, 22.58, 26.19, 25.71], dtype=np.float64)
    # Low tolerance because test values were truncated to 2 decimal places.
    assert_allclose(catchment1.flow, step1, atol=1e-1)

    model.step()
    step2 = np.array([20.03, 20.10, 22.18, 21.62, 20.17, 19.92, 19.63, 20.90, 24.24, 23.80], dtype=np.float64)
    # Low tolerance because test values were truncated to 2 decimal places.
    assert_allclose(catchment1.flow, step2, atol=1e-1)

    model.finish()


def test_timeseries_with_wrong_hash():
    with pytest.raises(HashMismatchError):
        load_model('timeseries2_hdf_wrong_hash.json')


def test_tablesarrayparameter_scenario_slice():
    model = load_model('timeseries2_hdf.json')
    catchment1 = model.nodes['catchment1']
    scenario = model.scenarios["scenario A"]
    scenario.slice = slice(0, 10, 2)
    model.setup()
    model.reset()
    model.step()
    step1 = np.array([21.64, 21.72, 23.97, 23.35, 21.79, 21.52, 21.21, 22.58, 26.19, 25.71], dtype=np.float64)
    assert_allclose(catchment1.flow, step1[::2], atol=1e-1)
    model.step()
    step2 = np.array([20.03, 20.10, 22.18, 21.62, 20.17, 19.92, 19.63, 20.90, 24.24, 23.80], dtype=np.float64)
    assert_allclose(catchment1.flow, step2[::2], atol=1e-1)
    model.finish()

def test_tablesarrayparameter_scenario_user_combinations():
    """Test TablesArrayParameter with user defined combination of scenarios"""
    model = load_model('timeseries2_hdf.json')
    catchment1 = model.nodes['catchment1']
    scenario = model.scenarios["scenario A"]
    scenario2 = Scenario(model, "scenario B", size=2)
    # combinations are intentially out of order and with duplicates
    model.scenarios.user_combinations = [(0, 0), (2, 0), (6, 1), (2, 1)]
    model.setup()
    model.reset()
    assert(len(model.scenarios.combinations) == 4)
    model.step()
    step1 = np.array([21.64, 21.72, 23.97, 23.35, 21.79, 21.52, 21.21, 22.58, 26.19, 25.71], dtype=np.float64)
    assert_allclose(catchment1.flow, step1[[0, 2, 6, 2]], atol=1e-1)
    model.step()
    step2 = np.array([20.03, 20.10, 22.18, 21.62, 20.17, 19.92, 19.63, 20.90, 24.24, 23.80], dtype=np.float64)
    assert_allclose(catchment1.flow, step2[[0, 2, 6, 2]], atol=1e-1)
    model.finish()

def test_tables_array_index_error():
    # check an exception is raised (before the model starts) if the length
    # of the data passed to a TablesArrayParameter is not long enough
    model = load_model('timeseries2_hdf.json')
    model.timestepper.start = "1920-01-01"
    model.timestepper.end = "1980-01-01"
    with pytest.raises(IndexError):
        model.run()
    assert(model.timestepper.current is None)

    # check the HDF5 file was closed, despite an exception being raised
    catchment1 = model.nodes['catchment1']
    print(type(catchment1))
    param = catchment1.max_flow
    assert(param.h5store is None)

def test_dirty_scenario(simple_linear_model):
    """Adding a scenario to a model makes it dirty"""
    model = simple_linear_model
    model.setup()
    assert(not model.dirty)
    scenario = Scenario(model, "test", size=42)
    assert(model.dirty)

def test_scenario_slices(simple_linear_model):
    """Test slicing of scenarios"""
    model = simple_linear_model

    # create two scenarios
    s1 = Scenario(model=model, name="A", size=20)
    s2 = Scenario(model=model, name="B", size=3)

    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 20 * 3)

    s1.slice = slice(0, None, 2)
    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 10 * 3)

    # check multiindex respects scenario slices
    index = model.scenarios.multiindex
    assert(len(index.levels) == 2)
    assert(len(index.levels[0]) == 10)
    assert(len(index.levels[1]) == 3)
    assert(len(index.codes) == 2)
    assert(len(index.codes[0]) == 10 * 3)
    assert(len(index.codes[1]) == 10 * 3)
    assert(index.names == ["A", "B"])

    s2.slice = slice(1, 3, 1)
    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 10 * 2)

    assert(combinations[0].global_id == 0)
    assert(tuple(combinations[0].indices) == (0, 1))

    assert(combinations[-1].global_id == 19)
    assert(tuple(combinations[-1].indices) == (18, 2))

    model.run()

    node = model.nodes["Input"]
    assert((len(combinations),) == node.flow.shape)

    s1.slice = None
    s2.slice = None
    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 20 * 3)

def test_scenario_user_combinations(simple_linear_model):
    model = simple_linear_model

    # create two scenarios
    s1 = Scenario(model=model, name="A", size=20)
    s2 = Scenario(model=model, name="B", size=3)

    model.scenarios.user_combinations = [[0, 1], [1, 1]]
    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 2)
    # ScenarioCollection.shape is simply the number of combinations
    assert model.scenarios.shape == (2, )

    # Test wrong number of dimensions
    with pytest.raises(ValueError):
        model.scenarios.user_combinations = [0, 1, 1, 1]

    # Test out of range index
    with pytest.raises(ValueError):
        model.scenarios.user_combinations = [[19, 3], [2, 2]]
    with pytest.raises(ValueError):
        model.scenarios.user_combinations = [[-1, 2], [2, 2]]

def test_scenario_slices_json():
    model = load_model('scenario_with_slices.json')
    scenarios = model.scenarios
    assert(len(scenarios) == 2)
    assert(scenarios["scenario A"].slice == slice(0, None, 2))
    assert(scenarios["scenario B"].slice == slice(0, 1, 1))
    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 5)

def test_scenario_slices_json():
    model = load_model('scenario_with_user_combinations.json')
    scenarios = model.scenarios
    assert(len(scenarios) == 2)
    combinations = model.scenarios.get_combinations()
    assert(len(combinations) == 3)
